# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import torch
from torch import nn
import numpy as np
from ultra.baselines.dqn.dqn.network import *
from smarts.core.agent import Agent
from ultra.utils.common import merge_discrete_action_spaces, to_3d_action, to_2d_action
import pathlib, os, copy
from ultra.baselines.dqn.dqn.network import DQNWithSocialEncoder
from ultra.baselines.dqn.dqn.explore import EpsilonExplore
from ultra.baselines.common.replay_buffer import ReplayBuffer
from ultra.baselines.common.social_vehicle_config import get_social_vehicle_configs
from ultra.baselines.common.yaml_loader import load_yaml
from ultra.baselines.common.baseline_state_preprocessor import BaselineStatePreprocessor


class DQNPolicy(Agent):
    lane_actions = ["keep_lane", "slow_down", "change_lane_left", "change_lane_right"]

    def __init__(
        self,
        policy_params=None,
        checkpoint_dir=None,
    ):
        self.policy_params = policy_params
        network_class = DQNWithSocialEncoder
        self.epsilon_obj = EpsilonExplore(1.0, 0.05, 100000)
        action_space_type = policy_params["action_space_type"]
        if action_space_type == "continuous":
            discrete_action_spaces = [
                np.asarray([-0.25, 0.0, 0.5, 0.75, 1.0]),
                np.asarray(
                    [-1.0, -0.75, -0.5, -0.25, -0.1, 0.0, 0.1, 0.25, 0.5, 0.75, 1.0]
                ),
            ]
        else:
            discrete_action_spaces = [[0], [1]]
        action_size = discrete_action_spaces
        self.merge_action_spaces = 0 if action_space_type == "continuous" else -1

        self.step_count = 0
        self.update_count = 0
        self.num_updates = 0
        self.current_sticky = 0
        self.current_iteration = 0

        lr = float(policy_params["lr"])
        seed = int(policy_params["seed"])
        self.train_step = int(policy_params["train_step"])
        self.target_update = float(policy_params["target_update"])
        self.device_name = "cuda:0" if torch.cuda.is_available() else "cpu"
        self.device = torch.device(self.device_name)
        self.warmup = int(policy_params["warmup"])
        self.gamma = float(policy_params["gamma"])
        self.batch_size = int(policy_params["batch_size"])
        self.use_ddqn = policy_params["use_ddqn"]
        self.sticky_actions = int(policy_params["sticky_actions"])
        prev_action_size = int(policy_params["prev_action_size"])
        self.prev_action = np.zeros(prev_action_size)
        self.action_size = prev_action_size

        if self.merge_action_spaces == 1:
            index2action, action2index = merge_discrete_action_spaces(*action_size)
            self.index2actions = [index2action]
            self.action2indexs = [action2index]
            self.num_actions = [len(self.index2actions)]
        elif self.merge_action_spaces == 0:
            self.index2actions = [
                merge_discrete_action_spaces([each])[0] for each in action_size
            ]
            self.action2indexs = [
                merge_discrete_action_spaces([each])[1] for each in action_size
            ]
            self.num_actions = [len(e) for e in action_size]
        else:
            index_to_actions = [
                e.tolist() if not isinstance(e, list) else e for e in action_size
            ]
            action_to_indexs = {
                str(k): v
                for k, v in zip(
                    index_to_actions, np.arange(len(index_to_actions)).astype(np.int)
                )
            }
            self.index2actions, self.action2indexs = (
                [index_to_actions],
                [action_to_indexs],
            )
            self.num_actions = [len(index_to_actions)]

        # state preprocessing
        self.social_policy_hidden_units = int(
            policy_params["social_vehicles"].get("social_policy_hidden_units", 0)
        )
        self.social_capacity = int(
            policy_params["social_vehicles"].get("social_capacity", 0)
        )
        self.observation_num_lookahead = int(
            policy_params.get("observation_num_lookahead", 0)
        )
        self.social_policy_init_std = int(
            policy_params["social_vehicles"].get("social_policy_init_std", 0)
        )
        self.num_social_features = int(
            policy_params["social_vehicles"].get("num_social_features", 0)
        )
        self.social_vehicle_config = get_social_vehicle_configs(
            **policy_params["social_vehicles"]
        )

        self.social_vehicle_encoder = self.social_vehicle_config["encoder"]
        self.state_description = BaselineStatePreprocessor.get_state_description(
            policy_params["social_vehicles"],
            policy_params["observation_num_lookahead"],
            prev_action_size,
        )
        self.social_feature_encoder_class = self.social_vehicle_encoder[
            "social_feature_encoder_class"
        ]
        self.social_feature_encoder_params = self.social_vehicle_encoder[
            "social_feature_encoder_params"
        ]

        self.checkpoint_dir = checkpoint_dir
        self.reset()

        torch.manual_seed(seed)
        network_params = {
            "state_size": self.state_size,
            "social_feature_encoder_class": self.social_feature_encoder_class,
            "social_feature_encoder_params": self.social_feature_encoder_params,
        }
        self.online_q_network = network_class(
            num_actions=self.num_actions,
            **(network_params if network_params else {}),
        ).to(self.device)
        self.target_q_network = network_class(
            num_actions=self.num_actions,
            **(network_params if network_params else {}),
        ).to(self.device)
        self.update_target_network()

        self.optimizers = torch.optim.Adam(
            params=self.online_q_network.parameters(), lr=lr
        )
        self.loss_func = nn.MSELoss(reduction="none")

        if self.checkpoint_dir:
            self.load(self.checkpoint_dir)

        self.action_space_type = "continuous"
        self.to_real_action = to_3d_action

        self.replay = ReplayBuffer(
            buffer_size=int(policy_params["replay_buffer"]["buffer_size"]),
            batch_size=int(policy_params["replay_buffer"]["batch_size"]),
            device_name=self.device_name,
        )

    def lane_action_to_index(self, state):
        state = state.copy()
        if (
            len(state["action"]) == 3
            and (state["action"] == np.asarray([0, 0, 0])).all()
        ):  # initial action
            state["action"] = np.asarray([0])
        else:
            state["action"] = self.lane_actions.index(state["action"])
        return state

    @property
    def state_size(self):
        # Adjusting state_size based on number of features (ego+social)
        size = sum(self.state_description["low_dim_states"].values())
        if self.social_feature_encoder_class:
            size += self.social_feature_encoder_class(
                **self.social_feature_encoder_params
            ).output_dim
        else:
            size += self.social_capacity * self.num_social_features
        # adding the previous action
        size += self.action_size
        return size

    def reset(self):
        self.eps_throttles = []
        self.eps_steers = []
        self.eps_step = 0
        self.current_sticky = 0

    def soft_update(self, target, src, tau):
        for target_param, param in zip(target.parameters(), src.parameters()):
            target_param.detach_()
            target_param.copy_(target_param * (1.0 - tau) + param * tau)

    def update_target_network(self):
        self.target_q_network.load_state_dict(self.online_q_network.state_dict().copy())

    def act(self, *args, **kwargs):
        if self.current_sticky == 0:
            self.action = self._act(*args, **kwargs)
        self.current_sticky = (self.current_sticky + 1) % self.sticky_actions
        self.current_iteration += 1
        return self.to_real_action(self.action)

    def _act(self, state, explore=True):
        epsilon = self.epsilon_obj.get_epsilon()
        if not explore or np.random.rand() > epsilon:
            state = copy.deepcopy(state)
            state["low_dim_states"] = np.float32(
                np.append(state["low_dim_states"], self.prev_action)
            )
            state["social_vehicles"] = (
                torch.from_numpy(state["social_vehicles"]).unsqueeze(0).to(self.device)
            )
            state["low_dim_states"] = (
                torch.from_numpy(state["low_dim_states"]).unsqueeze(0).to(self.device)
            )
            self.online_q_network.eval()
            with torch.no_grad():
                qs = self.online_q_network(state)
            qs = [q.data.cpu().numpy().flatten() for q in qs]
            # out_str = " || ".join(
            #     [
            #         " ".join(
            #             [
            #                 "{}: {:.4f}".format(index2action[j], q[j])
            #                 for j in range(num_action)
            #             ]
            #         )
            #         for index2action, q, num_action in zip(
            #             self.index2actions, qs, self.num_actions
            #         )
            #     ]
            # )
            # print(out_str)
            inds = [np.argmax(q) for q in qs]
        else:
            inds = [np.random.randint(num_action) for num_action in self.num_actions]
        action = []
        for j, ind in enumerate(inds):
            action.extend(self.index2actions[j][ind])
        self.epsilon_obj.step()
        self.eps_step += 1
        action = np.asarray(action)
        return action

    def save(self, model_dir):
        model_dir = pathlib.Path(model_dir)
        torch.save(self.online_q_network.state_dict(), model_dir / "online.pth")
        torch.save(self.target_q_network.state_dict(), model_dir / "target.pth")

    def load(self, model_dir, cpu=False):
        model_dir = pathlib.Path(model_dir)
        print("loading from :", model_dir)

        map_location = None
        if cpu:
            map_location = torch.device("cpu")
        self.online_q_network.load_state_dict(
            torch.load(model_dir / "online.pth", map_location=map_location)
        )
        self.target_q_network.load_state_dict(
            torch.load(model_dir / "target.pth", map_location=map_location)
        )
        print("Model loaded")

    def step(self, state, action, reward, next_state, done, info, others=None):
        # dont treat timeout as done equal to True
        max_steps_reached = info["logs"]["events"].reached_max_episode_steps
        if max_steps_reached:
            done = False
        if self.action_space_type == "continuous":
            action = to_2d_action(action)
            _action = (
                [[e] for e in action]
                if not self.merge_action_spaces
                else [action.tolist()]
            )
            action_index = np.asarray(
                [
                    action2index[str(e)]
                    for action2index, e in zip(self.action2indexs, _action)
                ]
            )
        else:
            action_index = self.lane_actions.index(action)
            action = action_index
        self.replay.add(
            state=state,
            action=action_index,
            reward=reward,
            next_state=next_state,
            done=done,
            others=others,
            social_capacity=self.social_capacity,
            observation_num_lookahead=self.observation_num_lookahead,
            social_vehicle_config=self.social_vehicle_config,
            prev_action=self.prev_action,
        )
        if (
            self.step_count % self.train_step == 0
            and len(self.replay) >= self.batch_size
            and (self.warmup is None or len(self.replay) >= self.warmup)
        ):
            out = self.learn()
            self.update_count += 1
        else:
            out = {}

        if self.target_update > 1 and self.step_count % self.target_update == 0:
            self.update_target_network()
        elif self.target_update < 1.0:
            self.soft_update(
                self.target_q_network, self.online_q_network, self.target_update
            )
        self.step_count += 1
        self.prev_action = action

        return out

    def learn(self):
        states, actions, rewards, next_states, dones, others = self.replay.sample(
            device=self.device
        )
        if not self.merge_action_spaces:
            actions = torch.chunk(actions, len(self.num_actions), -1)
        else:
            actions = [actions]

        self.target_q_network.eval()
        with torch.no_grad():
            qs_next_target = self.target_q_network(next_states)

        if self.use_ddqn:
            self.online_q_network.eval()
            with torch.no_grad():
                qs_next_online = self.online_q_network(next_states)
            next_actions = [
                torch.argmax(q_next_online, dim=1, keepdim=True)
                for q_next_online in qs_next_online
            ]
        else:
            next_actions = [
                torch.argmax(q_next_target, dim=1, keepdim=True)
                for q_next_target in qs_next_target
            ]

        qs_next_target = [
            torch.gather(q_next_target, 1, next_action)
            for q_next_target, next_action in zip(qs_next_target, next_actions)
        ]

        self.online_q_network.train()
        qs, aux_losses = self.online_q_network(states, training=True)
        qs = [torch.gather(q, 1, action.long()) for q, action in zip(qs, actions)]
        qs_target_value = [
            rewards + self.gamma * (1 - dones) * q_next_target
            for q_next_target in qs_next_target
        ]
        td_loss = [
            self.loss_func(q, q_target_value).mean()
            for q, q_target_value in zip(qs, qs_target_value)
        ]
        mean_td_loss = sum(td_loss) / len(td_loss)

        loss = mean_td_loss + sum(
            [e["value"] * e["weight"] for e in aux_losses.values()]
        )

        self.optimizers.zero_grad()
        loss.backward()
        self.optimizers.step()

        out = {}
        out.update(
            {
                "loss/td{}".format(j): {
                    "type": "scalar",
                    "data": td_loss[j].data.cpu().numpy(),
                    "freq": 10,
                }
                for j in range(len(td_loss))
            }
        )
        out.update(
            {
                "loss/{}".format(k): {
                    "type": "scalar",
                    "data": v["value"],  # .detach().cpu().numpy(),
                    "freq": 10,
                }
                for k, v in aux_losses.items()
            }
        )
        out.update({"loss/all": {"type": "scalar", "data": loss, "freq": 10}})

        self.num_updates += 1
        return out



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import argparse
import copy
import glob
import json
import os
import random
import re
import timeit
from collections import defaultdict

import dill
import gym
import numpy as np
import ray
import torch
from matplotlib import pyplot as plt

from smarts.core.agent import Agent, AgentSpec
from smarts.core.agent_interface import AgentInterface, AgentType, NeighborhoodVehicles
from smarts.core.controllers import ActionSpaceType
from smarts.core.utils.episodes import episodes
from ultra.scenarios.analysis.base_analysis import BaseAnalysis
from ultra.scenarios.common.social_vehicle_definitions import get_social_vehicle_color
from ultra.scenarios.common.visualization import (
    convert_to_gif,
    draw_intersection,
    profile_vehicles,
)

num_gpus = 1 if torch.cuda.is_available() else 0


@ray.remote(num_gpus=num_gpus / 2)
class ScenarioAnalysis(BaseAnalysis):
    def __init__(self):
        super(BaseAnalysis, self).__init__()
        self.analysis = {
            "stopwatcher_max_steps": [],
            "max_num_step_overall_routes": [],
            "steps_taken_by_behavior": defaultdict(lambda: []),
            "steps_taken_by_route": defaultdict(lambda: []),
            "stops_by_route": defaultdict(lambda: []),
            "episode_number_of_steps": [],
            "vehicle_states": [],
            "stop_percentage": defaultdict(lambda: []),
        }
        self.all_simulation_data = []

    def log_steps_summary(self, file_name):
        out_str = []
        all_vehicle_states = []
        for state in self.analysis["vehicle_states"]:
            all_vehicle_states.extend(list(state.values()))
        all_steps = []
        for state in all_vehicle_states:
            if state["route"] == ("south-SN", "west-EW"):
                print(state["end_step"] - state["start_step"])
                all_steps.append(state["end_step"] - state["start_step"])
        all_steps = np.asarray(all_steps)
        steps_thresholds = [100, 200, 500, 1000]
        percentage = [
            np.sum(all_steps > threshold) / len(all_steps)
            for threshold in steps_thresholds
        ]
        out_str.append(f"Step south-west threshold: \t{steps_thresholds}")
        out_str.append(f"Threshold south-west percentage: \t{percentage}")
        with open(file_name, "w") as f:
            f.writelines("\n".join(out_str))

    def analyze(self, episode_step, stopwatcher_max_steps, simulation_data):
        max_step = 0
        for v_id, v_state in self.social_vehicles_states.items():
            step_taken = v_state["end_step"] - v_state["start_step"]
            self.analysis["steps_taken_by_route"][v_state["route"]].append(step_taken)
            self.analysis["steps_taken_by_behavior"][v_state["behavior"]].append(
                step_taken
            )
            self.analysis["stops_by_route"][v_state["route"]].append(
                v_state["stop_step"]  # / time_taken
            )
            if v_state["route"] != ("south-SN", "west-EW"):
                max_time = max(max_step, step_taken)
        self.analysis["episode_number_of_steps"].append(episode_step)
        self.analysis["max_num_step_overall_routes"].append(max_step)
        self.analysis["stopwatcher_max_steps"].append(stopwatcher_max_steps)
        for v_r in self.analysis["stops_by_route"].keys():
            without_zeros = [i for i in self.analysis["stops_by_route"][v_r] if i != 0]
            percent = len(without_zeros) / len(self.analysis["stops_by_route"][v_r])
            self.analysis["stop_percentage"][v_r].append(percent)

        self.all_simulation_data.append(simulation_data)
        self.analysis["vehicle_states"].append(
            copy.deepcopy(self.social_vehicles_states)
        )

        # self.analysis['all_stop_percentage'] = defaultdict(lambda: [])
        # for episode_vs in self.analysis["vehicle_states"]:
        #     episode_stop_steps = defaultdict(lambda: [])
        #     for vs in episode_vs.values():
        #         episode_stop_steps[vs["route"]].append(vs["stop_steps"])
        #     for k, v in episode_stop_steps.items():
        #         without_zeros = [e for e in v if e > 0]
        #         self.analysis['all_stop_percentage'][k].append(len(without_zeros) / len(v))

    def draw_plots(self, save_dir, failure_time):
        self.generate_figures1(save_dir=f"{save_dir}/analysis1")
        self.generate_figures2(
            save_dir=f"{save_dir}/analysis2", failure_time=failure_time
        )
        self.log_steps_summary(f"{save_dir}/step_summary.txt")
        print(f"Plots/results saved in {save_dir}")

    def generate_figures1(self, save_dir):
        print("Generating Analysis 1")
        os.makedirs(save_dir, exist_ok=True)
        #         self.save_histogram(
        #             self.analysis["max_num_step_overall_routes"],
        #             f"{save_dir}/max_num_step_overall_routes.png",
        #             "max_num_step_overall_routes",
        #         )
        self.save_histogram(
            self.analysis["stopwatcher_max_steps"],
            f"{save_dir}/stopwatcher_max_steps.png",
            "stopwatcher_max_steps",
            (0, 6000),
        )

        self.save_histogram(
            self.analysis["episode_number_of_steps"],
            f"{save_dir}/step_episode.png",
            "time step_episode",
            (0, 6000),
        )

        for v_t in self.analysis["steps_taken_by_behavior"].keys():
            self.save_histogram(
                self.analysis["steps_taken_by_behavior"][v_t],
                f"{save_dir}/steps_taken_by_{v_t}.png",
                f"steps {v_t}",
                (0, 6000),
            )
        for v_r in self.analysis["steps_taken_by_route"].keys():
            self.save_histogram(
                self.analysis["steps_taken_by_route"][v_r],
                f"{save_dir}/steps_taken_by_route{v_r}.png",
                f"steps {v_r}",
                (0, 6000),
            )
        self.save_histogram(
            sum(list(self.analysis["steps_taken_by_behavior"].values()), []),
            f"{save_dir}/steps_taken_by_behavior.png",
            "steps all",
            (0, 6000),
        )
        for v_r in self.analysis["stops_by_route"].keys():
            without_zeros = [i for i in self.analysis["stops_by_route"][v_r] if i != 0]
            self.save_histogram(
                without_zeros,
                f"{save_dir}/stops_by_route_{v_r}.png",
                f"stop % {v_r}",
                (0, 6000),
            )
        self.save_histogram(
            sum(list(self.analysis["stops_by_route"].values()), []),
            f"{save_dir}/stop_by_routes_all.png",
            "stop % all",
            (0, 6000),
        )

        # temporary remove:
        if "all_stop_percentage" not in self.analysis:
            self.analysis["all_stop_percentage"] = defaultdict(lambda: [])
            for episode_vs in self.analysis["vehicle_states"]:
                episode_stop_steps = defaultdict(lambda: [])
                for vs in episode_vs.values():
                    episode_stop_steps[vs["route"]].append(vs["stop_step"])
                for k, v in episode_stop_steps.items():
                    without_zeros = [e for e in v if e > 0]
                    self.analysis["all_stop_percentage"][k].append(
                        len(without_zeros) / len(v)
                    )

        for v_r in self.analysis["all_stop_percentage"].keys():
            self.save_histogram(
                self.analysis["all_stop_percentage"][v_r],
                f"{save_dir}/stop_percentage{v_r}.png",
                f"stop percentage {v_r}",
                (0, 1),
            )

    def generate_figures2(self, save_dir, failure_time, num_bins=20):
        print("Generating Analysis 2")
        total_travel_time_list = []
        travel_percentile_list = []
        blocked_travel_time_list = []
        num_vehicle_list = []
        statistics_list = []
        failed_list = []
        total_south_west = 0
        solved_south_west = 0
        num_scenarios = len(self.all_simulation_data)

        for scenario in self.all_simulation_data:
            statistics = {"destination": {}, "route": {}}
            scenario_travel_time = []
            scenario_blocked_time = []
            for veh_id in scenario["vehicle_data"]:
                destination_key = scenario["vehicle_data"][veh_id]["destination"]
                route_key = scenario["vehicle_data"][veh_id]["route"]

                is_successful = scenario["vehicle_data"][veh_id]["success"]
                is_rerouted = not scenario["vehicle_data"][veh_id]["rerouted"] is None
                is_still_running = not is_successful and not is_rerouted
                travel_time = scenario["vehicle_data"][veh_id]["travel_time"]
                blocked_time = scenario["vehicle_data"][veh_id]["blocked_time"]
                has_failed = travel_time > failure_time or is_rerouted

                _from, _to = route_key.split("|")
                _from, _to = _from.split("-")[1], _to.split("-")[1]
                if _from == "south" and _to == "west":
                    total_south_west += 1
                    if travel_time < 101:
                        solved_south_west += 1

                if not destination_key in statistics["destination"]:
                    statistics["destination"][destination_key] = {
                        "blocked_time": [],
                        "travel_time": [],
                        "success": 0,
                        "rerouted": 0,
                        "in_simulation": 0,
                        "failure_time": 0,
                        "num_vehicles": 0,
                    }
                if not route_key in statistics["route"]:
                    statistics["route"][route_key] = {
                        "blocked_time": [],
                        "travel_time": [],
                        "success": 0,
                        "rerouted": 0,
                        "in_simulation": 0,
                        "failure_time": 0,
                        "num_vehicles": 0,
                    }

                statistics["destination"][destination_key]["success"] += int(
                    is_successful
                )
                statistics["destination"][destination_key]["rerouted"] += int(
                    is_rerouted
                )
                statistics["destination"][destination_key]["in_simulation"] += int(
                    is_still_running
                )
                statistics["destination"][destination_key]["failure_time"] += int(
                    has_failed
                )
                statistics["destination"][destination_key]["num_vehicles"] += 1

                statistics["destination"][destination_key]["travel_time"].append(
                    travel_time
                )
                statistics["destination"][destination_key]["blocked_time"].append(
                    blocked_time
                )

                statistics["route"][route_key]["success"] += int(is_successful)
                statistics["route"][route_key]["rerouted"] += int(is_rerouted)
                statistics["route"][route_key]["in_simulation"] += int(is_still_running)
                statistics["route"][route_key]["failure_time"] += int(has_failed)
                statistics["route"][route_key]["num_vehicles"] += 1

                statistics["route"][route_key]["travel_time"].append(travel_time)
                statistics["route"][route_key]["blocked_time"].append(blocked_time)

            travel_time_list = []
            blocked_time_list = []
            scenario_failed = 0
            for destination_key in statistics["destination"]:
                travel_time_list.extend(
                    statistics["destination"][destination_key]["travel_time"]
                )
                blocked_time_list.extend(
                    statistics["destination"][destination_key]["blocked_time"]
                )
                scenario_failed += statistics["destination"][destination_key][
                    "failure_time"
                ]

            scenario_99_percentile_travel_time = np.percentile(travel_time_list, 99)
            scenario_total_travel_time = np.sum(travel_time_list)
            scenario_blocked_time = np.sum(blocked_time_list)

            total_travel_time_list.append(scenario_total_travel_time)
            travel_percentile_list.append(scenario_99_percentile_travel_time)
            blocked_travel_time_list.append(scenario_blocked_time)

            failed_list.append(scenario_failed / max(len(scenario["vehicle_data"]), 1))
            num_vehicle_list.append(len(scenario["vehicle_data"]))
            statistics_list.append(statistics)

        # Append all scenario statistics together
        all_statistics = {"destination": {}, "route": {}}
        for scenario_statistics in statistics_list:
            for destination_key in scenario_statistics["destination"]:
                if not destination_key in all_statistics["destination"]:
                    all_statistics["destination"][destination_key] = {
                        "blocked_time": [],
                        "travel_time": [],
                        "success": [],
                        "rerouted": [],
                        "in_simulation": [],
                        "failure_time": [],
                        "num_vehicles": [],
                    }
                all_statistics["destination"][destination_key]["success"].append(
                    scenario_statistics["destination"][destination_key]["success"]
                )
                all_statistics["destination"][destination_key]["rerouted"].append(
                    scenario_statistics["destination"][destination_key]["rerouted"]
                )
                all_statistics["destination"][destination_key]["in_simulation"].append(
                    scenario_statistics["destination"][destination_key]["in_simulation"]
                )
                all_statistics["destination"][destination_key]["travel_time"].extend(
                    scenario_statistics["destination"][destination_key]["travel_time"]
                )
                all_statistics["destination"][destination_key]["blocked_time"].extend(
                    scenario_statistics["destination"][destination_key]["blocked_time"]
                )
                all_statistics["destination"][destination_key]["failure_time"].append(
                    scenario_statistics["destination"][destination_key]["failure_time"]
                    / max(
                        scenario_statistics["destination"][destination_key][
                            "num_vehicles"
                        ],
                        1,
                    )
                )
                all_statistics["destination"][destination_key]["num_vehicles"].append(
                    scenario_statistics["destination"][destination_key]["num_vehicles"]
                )

            for route_key in scenario_statistics["route"]:
                if not route_key in all_statistics["route"]:
                    all_statistics["route"][route_key] = {
                        "blocked_time": [],
                        "travel_time": [],
                        "success": [],
                        "rerouted": [],
                        "in_simulation": [],
                        "failure_time": [],
                        "num_vehicles": [],
                    }
                all_statistics["route"][route_key]["success"].append(
                    scenario_statistics["route"][route_key]["success"]
                )
                all_statistics["route"][route_key]["rerouted"].append(
                    scenario_statistics["route"][route_key]["rerouted"]
                )
                all_statistics["route"][route_key]["in_simulation"].append(
                    scenario_statistics["route"][route_key]["in_simulation"]
                )
                all_statistics["route"][route_key]["travel_time"].extend(
                    scenario_statistics["route"][route_key]["travel_time"]
                )
                all_statistics["route"][route_key]["blocked_time"].extend(
                    scenario_statistics["route"][route_key]["blocked_time"]
                )
                all_statistics["route"][route_key]["failure_time"].append(
                    scenario_statistics["route"][route_key]["failure_time"]
                    / max(scenario_statistics["route"][route_key]["num_vehicles"], 1)
                )
                all_statistics["route"][route_key]["num_vehicles"].append(
                    scenario_statistics["route"][route_key]["num_vehicles"]
                )

        # For each destination and route, we generate 5 histogram ....  + 2 for the travel time and percentile ...
        os.makedirs(save_dir, exist_ok=True)
        for destination_key in all_statistics["destination"]:
            sub_figure_dir = os.path.join(save_dir, destination_key)
            os.makedirs(sub_figure_dir, exist_ok=True)
            for statistics_key in all_statistics["destination"][destination_key]:
                self.save_histogram(
                    all_statistics["destination"][destination_key][statistics_key],
                    f"{sub_figure_dir}/{statistics_key}.png",
                    statistics_key,
                )
        for route_key in all_statistics["route"]:
            sub_figure_dir = os.path.join(save_dir, route_key)
            os.makedirs(sub_figure_dir, exist_ok=True)
            for statistics_key in all_statistics["route"][route_key]:
                self.save_histogram(
                    all_statistics["route"][route_key][statistics_key],
                    f"{sub_figure_dir}/{statistics_key}.png",
                    statistics_key,
                )

        self.save_histogram(
            total_travel_time_list,
            f"{save_dir}/total_travel_time.png",
            "total_travel_time",
        )

        self.save_histogram(
            travel_percentile_list,
            f"{save_dir}/99_percentile_travel_time.png",
            "99_percentile_travel_time",
        )

        self.save_histogram(
            blocked_travel_time_list,
            f"{save_dir}/blocked_travel_time.png",
            "blocked_travel_time",
        )

        self.save_histogram(
            num_vehicle_list,
            f"{save_dir}/num_vehicles.png",
            "num_vehicles",
        )

        self.save_histogram(
            failed_list,
            f"{save_dir}/failed_vehicles.png",
            "failed_vehicles",
        )

        if total_south_west > 0:
            print(
                "Solved south-west: {:.2f}".format(solved_south_west / total_south_west)
            )
        else:
            print("No vehicles moved from south to west.")

    def save_data(self, save_dir):
        with open(f"{save_dir}/analysis.pkl", "wb") as handle:
            dill.dump([self.analysis, self.all_simulation_data], handle)
        print(f"saved data to {save_dir}/analysis.pkl")

    def load_data(self, path):
        with open(path, "rb") as handle:
            self.analysis, self.all_simulation_data = dill.load(handle)
        print(f"loaded data {len(self.analysis)} from {path}")

    def run(self, **params):
        super().run(**params, analyze_func=self.analyze)



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import math
import os
import random
import shutil

import cv2
import dill
import numpy as np
import torch
from matplotlib import pyplot as plt
from scipy.spatial.distance import euclidean
from skimage.transform import resize

import ultra.utils.geometry as geometry
from scipy.spatial.distance import euclidean
import math, datetime


def gen_experiment_name():
    dt = datetime.datetime.today()
    return f"experiment-{dt.year}.{dt.month}.{dt.day}-{dt.hour}:{dt.minute}:{dt.second}"


def rotate2d_vector(vectors, angle):
    ae_cos = np.cos(angle)
    ae_sin = np.sin(angle)
    rot_matrix = np.array([[ae_cos, -ae_sin], [ae_sin, ae_cos]])

    vectors_rotated = np.inner(vectors, rot_matrix)
    return vectors_rotated


def plot_experiment_data(experiment_path, save_dir):
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
    with open(f"{experiment_path}/results.pkl", "rb") as handle:
        data = dill.load(handle)
        iterations = sorted(data.keys())
        plot_names = data[iterations[0]].keys()
        for plot_name in plot_names:
            plt.figure()
            values = []
            for itr in iterations:
                values.append(data[itr][plot_name])
            plt.plot(iterations, values)
            plt.savefig(f"{save_dir}/{plot_name}.png")
            plt.close()


def seed_all_libraries(seed, max_rd=100000):
    np.random.seed(seed)
    rd_seed, np_seed, torch_seed = (
        np.random.randint(1, max_rd),
        np.random.randint(1, max_rd),
        np.random.randint(1, max_rd),
    )
    random.seed(rd_seed)
    np.random.seed(np_seed)
    torch.cuda.manual_seed_all(torch_seed)


def to_3d_action(action):
    throttle, steering = action
    brake = 0.0
    if throttle < 0.0:
        brake = abs(throttle)
        throttle = 0.0
    return np.asarray([throttle, brake, steering])


def to_2d_action(action):
    throttle, brake, steering = action
    if throttle != 0:
        accel = throttle
    else:
        accel = -brake if brake > 0 else 0.0
    return np.asarray([accel, steering])


def merge_discrete_action_spaces(*discrete_action_spaces):
    r = np.meshgrid(*discrete_action_spaces)
    r = np.column_stack([each.flatten() for each in r])
    index_to_action = [e.tolist() for e in r]
    action_to_index = {
        str(k): v
        for k, v in zip(index_to_action, np.arange(len(index_to_action)).astype(np.int))
    }
    return index_to_action, action_to_index


def deg2rad(x):
    return x / 180.0 * np.pi


def copytree(src, dst, symlinks=False, ignore=None):
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            if os.path.exists(d):
                shutil.rmtree(d)
            shutil.copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)


def copy_source_code(srcs, dst):
    for s in srcs:
        d = os.path.join(dst, os.path.basename(s))
        if os.path.isdir(s):
            os.makedirs(d, exist_ok=True)
        (copytree if os.path.isdir(s) else shutil.copy)(s, d)


def resize_im(im, size):
    im = (resize(im, size, anti_aliasing=True, mode="constant") * 255.0).astype(
        np.uint8
    )
    return im


def normalize_im(x):
    # normalize from 0->255.0 to [-1, 1]
    x = x / 255.0
    x = 2.0 * x - 1
    return x


def get_box(vehicle, safety_dist):

    heading = vehicle.heading
    if heading > np.pi:
        heading = heading - 2 * np.pi

    del_dist = safety_dist / 2.0
    x_dist = -del_dist * np.sin(heading)
    y_dist = del_dist * np.cos(heading)

    return geometry.Box(
        width=vehicle.bounding_box.width,
        height=vehicle.bounding_box.length + safety_dist,
        length=vehicle.bounding_box.height,  # height of vehicle
        centerX=0.0,
        centerY=0.0,
        centerZ=0.0,
        yaw=0.0,
        roll=0.0,
        pitch=heading,  # regular yaw
        translationX=vehicle.position[0] + x_dist,
        translationY=vehicle.position[1] + y_dist,
        translationZ=vehicle.position[2],
    )


def ego_social_safety(
    agent_obs,
    d_min_ego=0.01,
    t_c_ego=0.5,
    d_min_social=0.01,
    t_c_social=0.5,
    ignore_vehicle_behind=False,
):
    # coordinate of this thing:
    # position: [x, y, z]
    # x increase when go from left to right
    # y increase when go from bottom to top
    # angle increase when vehicle move counterclockwise direction with 0 radian = bottom to top direction
    # so this is just standard coordinate shifted by pi/2 radian counterclockwise
    # add pi/2 radian to every heading to obtain the heading in standard coordinate

    def get_relative_position_vector_angle(v1, v2):
        x = v2.position[0] - v1.position[0]
        y = v2.position[1] - v1.position[1]
        # angle = clip_angle_to_pi(np.arctan2(y, x) + np.pi / 2)
        angle = clip_angle_to_pi(np.arctan2(x, y))
        return angle

    def is_behind(ego_heading_relative_diff, threshold=np.pi / 8):
        return abs(ego_heading_relative_diff) < threshold

    def get_vehicles_not_behind(ego, socials):
        ego_angle = ego.heading
        relative_position_vector_angles = [
            -get_relative_position_vector_angle(e, ego) for e in socials
        ]
        ego_heading_relative_diffs = [
            clip_angle_to_pi(ego_angle - e) for e in relative_position_vector_angles
        ]
        idxs = [
            e
            for e in range(len(socials))
            if not is_behind(ego_heading_relative_diffs[e])
        ]
        vehicles_not_behind = [socials[e] for e in idxs]
        return vehicles_not_behind

    # for debugging, you can use visualize_social_safety() in scenarios/visualization.py

    neighborhood_vehicle_states = agent_obs.neighborhood_vehicle_states
    ego_vehicle_state = agent_obs.ego_vehicle_state
    if ignore_vehicle_behind:
        neighborhood_vehicle_states = get_vehicles_not_behind(
            ego_vehicle_state, neighborhood_vehicle_states
        )

    vehicles_bounding_boxes = []
    vehicles_bounding_boxes_safety = []

    safety_dist_ego = d_min_ego + ego_vehicle_state.speed * t_c_ego

    ego_bounding_box_safety = get_box(ego_vehicle_state, safety_dist=safety_dist_ego)
    ego_bounding_box = get_box(ego_vehicle_state, safety_dist=0.0)

    for vehicle in neighborhood_vehicle_states:
        safety_dist_social = d_min_social + vehicle.speed * t_c_social
        vehicles_bounding_boxes.append(get_box(vehicle, safety_dist=0.0))
        vehicles_bounding_boxes_safety.append(
            get_box(vehicle, safety_dist=safety_dist_social)
        )

    ego_num_violations = 0
    social_num_violations = 0
    for vehicle, vehicle_safety in zip(
        vehicles_bounding_boxes, vehicles_bounding_boxes_safety
    ):
        if ego_bounding_box_safety.intersects(vehicle):
            ego_num_violations += 1
        if ego_bounding_box.intersects(vehicle_safety):
            social_num_violations += 1

    return ego_num_violations, social_num_violations


def get_closest_waypoint(ego_position, ego_heading, num_lookahead, goal_path):
    closest_wp = min(goal_path, key=lambda wp: wp.dist_to(ego_position))
    min_dist = float("inf")
    min_dist_idx = -1
    for i, wp in enumerate(goal_path):

        if wp.dist_to(ego_position) < min_dist:
            min_dist = wp.dist_to(ego_position)
            min_dist_idx = i
            closest_wp = wp

    waypoints_lookahead = [
        get_relative_pos(wp, ego_position)
        for wp in goal_path[
            min_dist_idx : min(min_dist_idx + num_lookahead, len(goal_path))
        ]
    ]
    if len(waypoints_lookahead) > 0:
        while len(waypoints_lookahead) < num_lookahead:
            waypoints_lookahead.append(waypoints_lookahead[-1])
    else:
        waypoints_lookahead = [
            get_relative_pos(closest_wp.pos, ego_position) for i in range(num_lookahead)
        ]

    waypoints_lookahead = rotate2d_vector(waypoints_lookahead, -ego_heading)
    return closest_wp, waypoints_lookahead


def get_relative_pos(waypoint, ego_pos):
    return [waypoint.pos[0] - ego_pos[0], waypoint.pos[1] - ego_pos[1]]


def get_dist_to_center(env_observation, closest_waypoint):
    ego = env_observation.ego_vehicle_state
    signed_dist_from_center = closest_waypoint.signed_lateral_error(ego.position)
    lane_hwidth = closest_waypoint.lane_width * 0.5
    norm_dist_from_center = signed_dist_from_center / lane_hwidth
    return norm_dist_from_center


def get_closest_point_index(pts_arr, pts):
    distance = [euclidean(each, pts) for each in pts_arr]
    return np.argmin(distance)


def get_path_to_goal(goal, paths, start):
    goal_pos = goal.position
    start_pos = start.position
    path_start_pts = [each[0].pos for each in paths]

    best_path_ind = get_closest_point_index(path_start_pts, start_pos)
    path = paths[best_path_ind]

    middle_point = path[int(len(path) / 2)]
    goal_lane_id = middle_point.lane_id
    goal_lane_index = middle_point.lane_index

    path_pts = [each.pos for each in path]
    end_path_ind = get_closest_point_index(path_pts, goal_pos)

    return path


def compute_sum_aux_losses(aux_losses):
    return sum([e["value"] * e["weight"] for e in aux_losses.values()])


def clip_angle_to_pi(angle):
    while angle < -np.pi:
        angle += np.pi * 2
    while angle > np.pi:
        angle -= np.pi * 2
    return angle


def compute_grad(x, prev_x, dt):
    grad = (x - prev_x) / dt if prev_x is not None else x / dt
    return grad



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import argparse
import os
import sys

if "SUMO_HOME" in os.environ:
    tools = os.path.join(os.environ["SUMO_HOME"], "tools")
    sys.path.append(tools)
else:
    sys.exit("please declare environment variable 'SUMO_HOME'")

import time

import matplotlib.pyplot as plt
import numpy as np
import traci
import traci.constants as tc

from ultra.scenarios.generate_scenarios import build_scenarios

# sumo -c  --remote-port=8813
# sumo -n map.net.xml -r traffic/all.rou.xml --remote-port=8813

# Aggregates information for each edge and lane ...
def edge_lane_data_function(traci, data_aggregator, **kwargs):
    if "lane_data" not in data_aggregator:
        data_aggregator["lane_data"] = {}
        for lane_id in traci.lane.getIDList():
            data_aggregator["lane_data"][lane_id] = {
                "num_vehicles": [],
                "occupancy": [],
            }
    if "edge_data" not in data_aggregator:
        data_aggregator["edge_data"] = {}
        for edge_id in traci.edge.getIDList():
            data_aggregator["edge_data"][edge_id] = {
                "num_vehicles": [],
                "occupancy": [],
            }

    for lane_id in traci.lane.getIDList():
        data_aggregator["lane_data"][lane_id]["num_vehicles"].append(
            traci.lane.getLastStepVehicleNumber(lane_id)
        )
        data_aggregator["lane_data"][lane_id]["occupancy"].append(
            traci.lane.getLastStepOccupancy(lane_id)
        )
    for edge_id in traci.edge.getIDList():
        data_aggregator["edge_data"][edge_id]["num_vehicles"].append(
            traci.edge.getLastStepVehicleNumber(edge_id)
        )
        data_aggregator["edge_data"][edge_id]["occupancy"].append(
            traci.edge.getLastStepOccupancy(edge_id)
        )


def vehicle_data_function(traci, data_aggregator, blocked_max_speed=0.1, **kwargs):
    vehicle_list = traci.vehicle.getIDList()
    if "vehicle_data" not in data_aggregator:
        data_aggregator["vehicle_data"] = {}

    if "destination" not in data_aggregator:
        data_aggregator["destination"] = {}

    if "route" not in data_aggregator:
        data_aggregator["route"] = {}

    if "vehicle_density" not in data_aggregator:
        data_aggregator["vehicle_density"] = {"All": {}}

    for veh_id in vehicle_list:
        if veh_id not in data_aggregator["vehicle_data"]:
            veh_route = traci.vehicle.getRoute(veh_id)
            data_aggregator["vehicle_density"]["All"][kwargs["last_step"]] = len(
                vehicle_list
            )
            data_aggregator["vehicle_data"][veh_id] = {
                "min_gap": traci.vehicle.getMinGap(veh_id),
                "final_target": veh_route[-1],
                "origin": veh_route[0],
                "destination": veh_route[-1],
                "success": False,
                "route": veh_route[0] + "|" + veh_route[-1],
                "rerouted": None,
                "travel_time": 0,
                "blocked_time": 0,
                "accumulated_blocked_time": 0,
                "positions": [],
                "lane_list": [],
                "last_step": kwargs["last_step"],
            }
            if veh_route[-1] not in data_aggregator["destination"]:
                data_aggregator["destination"][veh_route[-1]] = 1
            else:
                data_aggregator["destination"][veh_route[-1]] += 1

            if veh_route[0] + "|" + veh_route[-1] not in data_aggregator["route"]:
                data_aggregator["route"][veh_route[0] + "|" + veh_route[-1]] = 1
            else:
                data_aggregator["route"][veh_route[0] + "|" + veh_route[-1]] += 1
        else:
            data_aggregator["vehicle_data"][veh_id]["travel_time"] += 1
            data_aggregator["vehicle_data"][veh_id]["positions"].append(
                traci.vehicle.getPosition(veh_id)
            )

            if traci.vehicle.getSpeed(veh_id) < blocked_max_speed:
                data_aggregator["vehicle_data"][veh_id]["blocked_time"] += 1
                if traci.vehicle.getLeader(veh_id) is None:
                    data_aggregator["vehicle_data"][veh_id][
                        "accumulated_blocked_time"
                    ] += 1
            else:
                data_aggregator["vehicle_data"][veh_id]["accumulated_blocked_time"] = 0

        if (
            data_aggregator["vehicle_data"][veh_id]["route"]
            not in data_aggregator["vehicle_density"]
        ):
            data_aggregator["vehicle_density"][
                data_aggregator["vehicle_data"][veh_id]["route"]
            ] = {}
        else:
            if (
                not kwargs["last_step"]
                in data_aggregator["vehicle_density"][
                    data_aggregator["vehicle_data"][veh_id]["route"]
                ]
            ):
                data_aggregator["vehicle_density"][
                    data_aggregator["vehicle_data"][veh_id]["route"]
                ][kwargs["last_step"]] = 0
            else:
                data_aggregator["vehicle_density"][
                    data_aggregator["vehicle_data"][veh_id]["route"]
                ][kwargs["last_step"]] += 1

        current_lane = traci.vehicle.getLaneID(veh_id)
        if not current_lane in data_aggregator["vehicle_data"][veh_id]["lane_list"]:
            data_aggregator["vehicle_data"][veh_id]["lane_list"].append(current_lane)
        data_aggregator["vehicle_data"][veh_id][
            "current_edge"
        ] = traci.vehicle.getRoadID(veh_id)
        if (
            data_aggregator["vehicle_data"][veh_id]["current_edge"]
            == data_aggregator["vehicle_data"][veh_id]["final_target"]
        ):
            data_aggregator["vehicle_data"][veh_id]["success"] = True


def sumo_rerouting_routine(
    traci, data_aggregator, time_elapsed_until_reroute=20, **kwargs
):
    vehicle_list = traci.vehicle.getIDList()
    for veh_id in vehicle_list:
        vehicle_lane_index = traci.vehicle.getLaneIndex(veh_id)
        bestLanes = traci.vehicle.getBestLanes(veh_id)
        if not bestLanes[vehicle_lane_index][4] and len(bestLanes) > 1:

            selected_lanes = []
            for j in range(len(bestLanes)):
                if j != vehicle_lane_index:
                    selected_lanes.append(j)

            select_lane = np.random.choice(selected_lanes)
            if select_lane > vehicle_lane_index:
                direction = 1
            else:
                direction = -1

            # Is changing lane, locked, and is leader of the lane -> Reroute
            if (
                not traci.vehicle.couldChangeLane(veh_id, direction)
                and data_aggregator["vehicle_data"][veh_id]["accumulated_blocked_time"]
                > time_elapsed_until_reroute
                and traci.vehicle.getLeader(veh_id) is None
            ):
                origin = data_aggregator["vehicle_data"][veh_id]["origin"].split("-")[1]
                final_target = data_aggregator["vehicle_data"][veh_id][
                    "final_target"
                ].split("-")[1]
                # print(origin, final_target)
                potential_list = []
                for potential_reroute_target in data_aggregator["destination"]:
                    token = potential_reroute_target.split("-")[1]
                    if token != final_target and token != origin:
                        potential_list.append(potential_reroute_target)
                result = np.random.choice(potential_list)
                try:
                    # traci.vehicle.changeTarget(veh_id, result)
                    new_route = list(
                        traci.simulation.findRoute(
                            data_aggregator["vehicle_data"][veh_id]["current_edge"],
                            result,
                        ).edges
                    )
                    traci.vehicle.setRoute(veh_id, new_route)
                    print("Route Change Success")
                    data_aggregator["vehicle_data"][veh_id][
                        "accumulated_blocked_time"
                    ] = 0
                    data_aggregator["vehicle_data"][veh_id]["rerouted"] = result
                except:
                    print("Route Change Failed")


def generate_figures(
    data,
    figure_dir,
    setting_id,
    failure_rate,
    num_steps,
    num_bins=50,
    comparisons=[[("route", "avg_blocked_time"), ("route", "failure_rate")]],
):
    # default_statistics = {"blocked_time":[], "travel_time":[], "success":0, "rerouted":0, "in_simulation":0}
    total_travel_time_list = []
    travel_percentile_list = []
    blocked_travel_time_list = []
    num_vehicle_list = []
    statistics_list = []
    failed_list = []
    num_scenarios = len(data)
    num_bins = min(num_scenarios // 2, num_bins)
    for scenario in data:
        statistics = {
            "destination": {},
            "route": {},
            "vehicle_gen_prob": scenario["vehicle_gen_prob"],
        }
        scenario_travel_time = []
        scenario_blocked_time = []
        for veh_id in scenario["vehicle_data"]:
            destination_key = scenario["vehicle_data"][veh_id]["destination"]
            route_key = scenario["vehicle_data"][veh_id]["route"]

            is_successful = scenario["vehicle_data"][veh_id]["success"]
            is_rerouted = not scenario["vehicle_data"][veh_id]["rerouted"] is None
            is_still_running = not is_successful and not is_rerouted
            travel_time = scenario["vehicle_data"][veh_id]["travel_time"]
            blocked_time = scenario["vehicle_data"][veh_id]["blocked_time"]
            has_failed = travel_time > failure_rate or is_rerouted

            if not destination_key in statistics["destination"]:
                statistics["destination"][destination_key] = {
                    "blocked_time": [],
                    "travel_time": [],
                    "success": 0,
                    "rerouted": 0,
                    "in_simulation": 0,
                    "deadlock_rate": 0,
                    "num_vehicles": 0,
                }
            if not route_key in statistics["route"]:
                statistics["route"][route_key] = {
                    "blocked_time": [],
                    "travel_time": [],
                    "success": 0,
                    "rerouted": 0,
                    "in_simulation": 0,
                    "deadlock_rate": 0,
                    "num_vehicles": 0,
                }

            statistics["destination"][destination_key]["success"] += int(is_successful)
            statistics["destination"][destination_key]["rerouted"] += int(is_rerouted)
            statistics["destination"][destination_key]["in_simulation"] += int(
                is_still_running
            )
            statistics["destination"][destination_key]["deadlock_rate"] += int(
                has_failed
            )
            statistics["destination"][destination_key]["num_vehicles"] += 1

            statistics["destination"][destination_key]["travel_time"].append(
                travel_time
            )
            statistics["destination"][destination_key]["blocked_time"].append(
                blocked_time
            )

            statistics["route"][route_key]["success"] += int(is_successful)
            statistics["route"][route_key]["rerouted"] += int(is_rerouted)
            statistics["route"][route_key]["in_simulation"] += int(is_still_running)
            statistics["route"][route_key]["deadlock_rate"] += int(has_failed)
            statistics["route"][route_key]["num_vehicles"] += 1

            statistics["route"][route_key]["travel_time"].append(travel_time)
            statistics["route"][route_key]["blocked_time"].append(blocked_time)

        travel_time_list = []
        blocked_time_list = []
        scenario_failed = 0
        for destination_key in statistics["destination"]:
            travel_time_list.extend(
                statistics["destination"][destination_key]["travel_time"]
            )
            blocked_time_list.extend(
                statistics["destination"][destination_key]["blocked_time"]
            )
            scenario_failed += statistics["destination"][destination_key][
                "deadlock_rate"
            ]

        scenario_99_percentile_travel_time = np.percentile(travel_time_list, 99)
        scenario_total_travel_time = np.mean(travel_time_list)
        scenario_blocked_time = np.mean(blocked_time_list)

        total_travel_time_list.append(scenario_total_travel_time)
        travel_percentile_list.append(scenario_99_percentile_travel_time)
        blocked_travel_time_list.append(scenario_blocked_time)

        failed_list.append(scenario_failed / max(len(scenario["vehicle_data"]), 1))
        num_vehicle_list.append(len(scenario["vehicle_data"]))
        statistics_list.append(statistics)

    index = 0
    for scenario in data:
        statistics = statistics_list[index]
        scenario_statistics = {}
        route_list = list(scenario["vehicle_density"].keys())
        for route_id in route_list:
            scenario_statistics[route_id] = []
        for i in range(num_steps):
            for route_id in route_list:
                if not i in scenario["vehicle_density"][route_id]:
                    scenario["vehicle_density"][route_id][i] = 0
                scenario_statistics[route_id].append(
                    scenario["vehicle_density"][route_id][i]
                )
        for route_id in route_list:
            scenario_statistics[route_id] = np.mean(scenario_statistics[route_id])
        statistics["vehicle_density"] = scenario_statistics
        index += 1

    # Append all scenario statistics together
    all_statistics = {
        "destination": {},
        "route": {},
        "vehicle_density": {},
        "vehicle_gen_prob": {},
    }
    all_statistics["route"]["All"] = {
        "avg_blocked_time": blocked_travel_time_list,
        "avg_travel_time": total_travel_time_list,
        "deadlock_rate": failed_list,
        "num_vehicles": num_vehicle_list,
    }

    for scenario_statistics in statistics_list:
        for destination_key in scenario_statistics["destination"]:
            if not destination_key in all_statistics["destination"]:
                all_statistics["destination"][destination_key] = {
                    "blocked_time": [],
                    "travel_time": [],
                    "success": [],
                    "rerouted": [],
                    "in_simulation": [],
                    "deadlock_rate": [],
                    "num_vehicles": [],
                }
            all_statistics["destination"][destination_key]["success"].append(
                scenario_statistics["destination"][destination_key]["success"]
            )
            all_statistics["destination"][destination_key]["rerouted"].append(
                scenario_statistics["destination"][destination_key]["rerouted"]
            )
            all_statistics["destination"][destination_key]["in_simulation"].append(
                scenario_statistics["destination"][destination_key]["in_simulation"]
            )
            all_statistics["destination"][destination_key]["travel_time"].extend(
                scenario_statistics["destination"][destination_key]["travel_time"]
            )
            all_statistics["destination"][destination_key]["blocked_time"].extend(
                scenario_statistics["destination"][destination_key]["blocked_time"]
            )

            all_statistics["destination"][destination_key]["deadlock_rate"].append(
                scenario_statistics["destination"][destination_key]["deadlock_rate"]
                / max(
                    scenario_statistics["destination"][destination_key]["num_vehicles"],
                    1,
                )
            )
            all_statistics["destination"][destination_key]["num_vehicles"].append(
                scenario_statistics["destination"][destination_key]["num_vehicles"]
            )

        for route_key in scenario_statistics["route"]:
            if not route_key in all_statistics["route"]:
                all_statistics["route"][route_key] = {
                    "blocked_time": [],
                    "travel_time": [],
                    "avg_travel_time": [],
                    "avg_blocked_time": [],
                    "success": [],
                    "rerouted": [],
                    "in_simulation": [],
                    "deadlock_rate": [],
                    "num_vehicles": [],
                }
            all_statistics["route"][route_key]["success"].append(
                scenario_statistics["route"][route_key]["success"]
            )
            all_statistics["route"][route_key]["rerouted"].append(
                scenario_statistics["route"][route_key]["rerouted"]
            )
            all_statistics["route"][route_key]["in_simulation"].append(
                scenario_statistics["route"][route_key]["in_simulation"]
            )
            all_statistics["route"][route_key]["travel_time"].extend(
                scenario_statistics["route"][route_key]["travel_time"]
            )
            all_statistics["route"][route_key]["blocked_time"].extend(
                scenario_statistics["route"][route_key]["blocked_time"]
            )

            all_statistics["route"][route_key]["avg_travel_time"].append(
                np.mean(scenario_statistics["route"][route_key]["travel_time"])
            )
            all_statistics["route"][route_key]["avg_blocked_time"].append(
                np.mean(scenario_statistics["route"][route_key]["blocked_time"])
            )

            all_statistics["route"][route_key]["deadlock_rate"].append(
                scenario_statistics["route"][route_key]["deadlock_rate"]
                / max(scenario_statistics["route"][route_key]["num_vehicles"], 1)
            )
            all_statistics["route"][route_key]["num_vehicles"].append(
                scenario_statistics["route"][route_key]["num_vehicles"]
            )
        for route_key in scenario_statistics["vehicle_density"]:
            if route_key not in all_statistics["vehicle_density"]:
                all_statistics["vehicle_density"][route_key] = {"vehicle_density": []}
            all_statistics["vehicle_density"][route_key]["vehicle_density"].append(
                scenario_statistics["vehicle_density"][route_key]
            )

        for route_key in scenario_statistics["vehicle_gen_prob"]:
            if route_key not in all_statistics["vehicle_gen_prob"]:
                all_statistics["vehicle_gen_prob"][route_key] = {"vehicle_gen_prob": []}
            all_statistics["vehicle_gen_prob"][route_key]["vehicle_gen_prob"].append(
                scenario_statistics["vehicle_gen_prob"][route_key]
            )
        # For each destination and route, we generate 5 histogram ....  + 2 for the travel time and percentile ...
    try:
        os.mkdir(figure_dir)
    except:
        print("Directories already exists ... ")
    figure_dir = figure_dir + "/" + setting_id
    try:
        os.mkdir(figure_dir)
    except:
        print("Directories already exists ... ")

    for overarching_key in all_statistics:
        for data_key in all_statistics[overarching_key]:
            sub_figure_dir = figure_dir + "/" + data_key
            try:
                os.mkdir(sub_figure_dir)
            except:
                print("Directories already exists ... ")
            for statistics_key in all_statistics[overarching_key][data_key]:
                n, bins, patches = plt.hist(
                    all_statistics[overarching_key][data_key][statistics_key],
                    num_bins,
                    facecolor="blue",
                )
                plt.title(
                    "Histogram Edge/Route_" + str(data_key) + "_" + str(statistics_key)
                )
                plt.xlabel(statistics_key)
                plt.ylabel(
                    "Number of occurences for " + str(num_scenarios) + " scenarios"
                )
                plt.savefig(sub_figure_dir + "/" + statistics_key + ".png")
                plt.close()

    # collect key informations
    for i in range(len(comparisons)):
        comparison = comparisons[i]
        key_00 = comparison[0][0]
        key_10 = comparison[1][0]
        key_01 = comparison[0][1]
        key_11 = comparison[1][1]
        print(comparison)
        id_list = list(all_statistics[key_00])
        for key_id in id_list:
            set_1 = all_statistics[key_00][key_id][key_01]
            set_2 = all_statistics[key_10][key_id][key_11]
            if len(set_1) == len(set_2):
                plt.plot(set_1, set_2, ".")
                plt.title(key_01 + " vs " + key_11)
                plt.xlabel(key_01)
                plt.ylabel(key_11)
                plt.savefig(figure_dir + "/" + key_id + "/" + key_01 + " vs " + key_11)
                plt.close()
            else:
                print("Invalid Comparison")


def sumo_traci_runner(
    simulation_time,
    custom_traci_functions,
    num_iterations,
    scenario_dir,
    simulation_step_size=1,
    init_time_skip=50,
    seed=0,
    run_per_scenario=2,
    **kwargs,
):
    skip = int(1 / simulation_step_size)
    np.random.seed(seed)
    all_simulation_data = []
    traci.init(8813)
    task = kwargs["task"]
    level = kwargs["level"]
    print(task, level)

    def dynamic_vehicle_gen_probability_func(pattern, current_step):
        if kwargs["vehicle_gen_prob"] > 0:
            current_step = current_step // run_per_scenario
            for edge_id in pattern["routes"]:
                if pattern["routes"][edge_id] is None:
                    continue
                pattern["routes"][edge_id]["begin_time_init"]["params"][
                    "probability"
                ] = (
                    kwargs["vehicle_gen_prob"]
                    + current_step * kwargs["dynamic_vehicle_gen_prob"]
                )
        return pattern

    build_scenarios(
        task=f"task{task}",
        level_name=level,
        num_seeds=num_iterations * run_per_scenario,
        has_stopwatcher=False,
        dynamic_pattern_func=dynamic_vehicle_gen_probability_func,
    )
    for j in range(num_iterations * run_per_scenario):
        current_time = time.time()
        current_scenario_dir = scenario_dir + "-flow-" + str(j)
        current_vehicle_gen_prob = (
            kwargs["vehicle_gen_prob"]
            + j // run_per_scenario * kwargs["dynamic_vehicle_gen_prob"]
        )
        # for k in range(run_per_scenario):
        #     # Reset scenario
        traci.load(
            [
                "-n",
                current_scenario_dir + "/map.net.xml",
                "-r",
                current_scenario_dir + "/traffic/all.rou.xml",
                "--seed",
                str(np.random.randint(0, run_per_scenario)),
            ]
        )
        simulation_data = {}  # Dictionary object to collect data ...
        accumulated_time = 0
        for i in range(int((simulation_time + init_time_skip) / simulation_step_size)):
            accumulated_time += simulation_step_size
            traci.simulationStep(accumulated_time)  # One step to initialize everything
            if i > int(init_time_skip / simulation_step_size) and i % skip == 0:
                # print("Sim:", j, "Sim Seed:", k,"Time Step:", traci.simulation.getTime(), "Num Vehicles:", len(traci.vehicle.getIDList()))
                for func in custom_traci_functions:
                    func(
                        traci,
                        simulation_data,
                        last_step=i - init_time_skip - 1,
                        **kwargs,
                    )

        simulation_data["vehicle_gen_prob"] = {"All": current_vehicle_gen_prob}
        for route_id in simulation_data["route"]:
            simulation_data["vehicle_gen_prob"][route_id] = current_vehicle_gen_prob

            all_simulation_data.append(simulation_data)
        print(
            "Sim:",
            j,
            "with",
            run_per_scenario,
            "seeds, completed with",
            time.time() - current_time,
            "seconds",
        )
    traci.close()
    # Reached End of simulation ...
    return all_simulation_data


# Default sumo instruction for 4 lane_t
# sumo -n research/edmonton/intersections/scenarios/task1/hard_4lane_t_80kmh_heavy_traffic_stress_test_t_intersection-mission-0-flow-0/map.net.xml -r research/edmonton/intersections/scenarios/task1/hard_4lane_t_80kmh_heavy_traffic_stress_test_t_intersection-mission-0-flow-0/traffic/all.rou.xml --remote-port=8813
# Run command example
# python research/edmonton/intersections/scenarios/sumo_experiment.py --level="hard" --task="1" --scenario_dir="research/edmonton/intersections/scenarios/task1/hard_4lane_t_80kmh_heavy_traffic_stress_test_t_intersection-mission-0" --figure_dir="research/edmonton/intersections/scenarios/task1/figures" --setting_id="hard_t_intersection_4_vehicle_prob_0_05" --num_scenario=100
if __name__ == "__main__":
    parser = argparse.ArgumentParser("generate scenarios")
    parser.add_argument("--task", help="type a task id [0, 1, 2, 3]", type=str)
    parser.add_argument("--level", help="easy/medium/hard", type=str)
    parser.add_argument("--scenario_dir", help="path", type=str)
    parser.add_argument("--figure_dir", help="path", type=str)
    parser.add_argument(
        "--dynamic_vehicle_gen_prob", help="path", type=float, default=0
    )
    parser.add_argument("--vehicle_gen_prob", help="path", type=float, default=0.05)
    parser.add_argument("--run_per_scenario", help="path", type=int, default=3)

    parser.add_argument(
        "--setting_id",
        help="for identifying figures",
        type=str,
        default="hard_t_intersection_0_05",
    )
    parser.add_argument("--reroute", help="path", type=str, default="False")
    parser.add_argument("--num_scenario", help="path", type=int, default=5)
    parser.add_argument("--scenario_run_time", help="path", type=int, default=1200)
    parser.add_argument("--failure_time", help="path", type=int, default=600)

    args = parser.parse_args()
    if args.level not in [
        "easy",
        "medium",
        "hard",
    ]:
        raise ValueError('level not accepted select from ["easy", "medium", "hard"]')
    if args.reroute == "True":
        sim_funcs = [
            edge_lane_data_function,
            vehicle_data_function,
            sumo_rerouting_routine,
        ]
    else:
        sim_funcs = [edge_lane_data_function, vehicle_data_function]
    data = sumo_traci_runner(
        args.scenario_run_time,
        sim_funcs,
        args.num_scenario,
        args.scenario_dir,
        task=args.task,
        level=args.level,
        vehicle_gen_prob=args.vehicle_gen_prob,
        dynamic_vehicle_gen_prob=args.dynamic_vehicle_gen_prob,
        run_per_scenario=args.run_per_scenario,
    )
    comparisons = [
        [("route", "avg_blocked_time"), ("route", "deadlock_rate")],
        [("vehicle_density", "vehicle_density"), ("route", "deadlock_rate")],
        [("vehicle_density", "vehicle_density"), ("route", "avg_travel_time")],
        [("vehicle_density", "vehicle_density"), ("route", "avg_blocked_time")],
        [("vehicle_gen_prob", "vehicle_gen_prob"), ("route", "avg_blocked_time")],
        [("vehicle_gen_prob", "vehicle_gen_prob"), ("route", "deadlock_rate")],
        [
            ("vehicle_gen_prob", "vehicle_gen_prob"),
            ("vehicle_density", "vehicle_density"),
        ],
    ]
    generate_figures(
        data,
        args.figure_dir,
        args.setting_id
        + "_reroute_"
        + str(args.reroute)
        + "_failure_time_"
        + str(args.failure_time),
        args.failure_time,
        args.scenario_run_time,
        comparisons=comparisons,
    )



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import argparse
import glob
import json
import os
import sys
import timeit
from collections import defaultdict

import dill
import ray

from smarts.core.agent import Agent, AgentSpec
from smarts.core.agent_interface import AgentInterface, NeighborhoodVehicles
from smarts.core.controllers import ActionSpaceType
from ultra.scenarios.analysis.behavior_analysis import BehaviorAnalysis
from ultra.scenarios.analysis.scenario_analysis import ScenarioAnalysis
from ultra.scenarios.analysis.sumo_experiment import (
    edge_lane_data_function,
    sumo_rerouting_routine,
    vehicle_data_function,
)
from ultra.scenarios.generate_scenarios import build_scenarios

if __name__ == "__main__":
    parser = argparse.ArgumentParser("scenarios")
    subparsers = parser.add_subparsers(help="sub-command help")

    parser_generate_scenarios = subparsers.add_parser(
        "generate", help="Generating scenarios"
    )
    parser_generate_scenarios.add_argument(
        "--task", help="type a task id [0, 1, 2, 3]", type=str
    )
    parser_generate_scenarios.add_argument(
        "--save-dir", help="directory for saving maps", type=str, default=None
    )
    parser_generate_scenarios.add_argument(
        "--root-dir",
        help="directory for saving maps",
        type=str,
        default="ultra/scenarios",
    )
    parser_generate_scenarios.add_argument("--level", help="easy/medium/hard", type=str)
    parser_generate_scenarios.add_argument(
        "--stopwatcher",
        help="aggressive/default/slow/blocker/crusher south-west",
        nargs="+",
    )

    parser_generate_scenarios.set_defaults(which="generate")

    parser_analyze_scenarios = subparsers.add_parser(
        "analyze", help="Analyzing scenarios"
    )
    parser_analyze_scenarios.add_argument(
        "--scenarios", help="type a task id [0, 1, 2, 3]", type=str
    )
    parser_analyze_scenarios.add_argument("--output", help="easy/medium/hard", type=str)
    parser_analyze_scenarios.add_argument("--failure-time", type=int, default=600)
    parser_analyze_scenarios.add_argument("--max-steps", type=int, default=6000)
    parser_analyze_scenarios.add_argument(
        "--video",
        help="record video by rate (default every 100 scenarios)",
        type=int,
        default=1000,
    )
    parser_analyze_scenarios.add_argument("--ego", help="path to the trained agent")
    parser_analyze_scenarios.add_argument("--policy", help="ego agent policy")
    parser_analyze_scenarios.add_argument(
        "--end-by-stopwatcher",
        help="end episode as stopwatcher exits",
        action="store_true",
    )
    parser_analyze_scenarios.set_defaults(which="analyze")

    parser_profile_vehicles = subparsers.add_parser("behavior", help="Profile vehicles")
    parser_profile_vehicles.add_argument(
        "--scenarios", help="type a task id [0, 1, 2, 3]", type=str
    )
    parser_profile_vehicles.add_argument(
        "--video",
        help="record video by rate (default every 100 scenarios)",
        type=int,
        default=1000,
    )
    parser_profile_vehicles.add_argument(
        "--end-by-stopwatcher",
        help="end episode as stopwatcher exits",
        action="store_true",
    )
    parser_profile_vehicles.add_argument("--max-steps", type=int, default=6000)
    parser_profile_vehicles.add_argument("--output", help="easy/medium/hard", type=str)
    parser_profile_vehicles.set_defaults(which="behavior")

    parser_plot_data = subparsers.add_parser(
        "plot-analysis", help="Plot existing analysis data"
    )
    parser_plot_data.add_argument("--data", help="path to the load data", type=str)
    parser_plot_data.add_argument("--output", help="path to save data", type=str)
    parser_plot_data.add_argument("--failure_time", type=int, default=600)
    parser_plot_data.set_defaults(which="plot-analysis")

    args = parser.parse_args()
    if args.which == "generate":
        stopwatcher_behavior, stopwatcher_route = None, None
        if args.stopwatcher:
            stopwatcher_behavior, stopwatcher_route = args.stopwatcher
        build_scenarios(
            task=f"task{args.task}",
            level_name=args.level,
            stopwatcher_behavior=stopwatcher_behavior,
            stopwatcher_route=stopwatcher_route,
            save_dir=args.save_dir,
            root_path=args.root_dir,
        )
    else:
        ray.init()
        if args.which == "plot-analysis":
            analyzer = ScenarioAnalysis.remote()
            ray.wait([analyzer.load_data.remote(args.data)])
            ray.wait(
                [
                    analyzer.draw_plots.remote(
                        save_dir=args.output, failure_time=args.failure_time
                    )
                ]
            )
        else:
            start = timeit.default_timer()
            if args.which == "analyze":
                analyzer = ScenarioAnalysis.remote()
            elif args.which == "behavior":
                analyzer = BehaviorAnalysis.remote()
                args.ego, args.policy, args.failure_time = None, None, None

            scenarios = glob.glob(f"{args.scenarios}*")
            timestep_sec = 0.1
            os.makedirs(args.output, exist_ok=True)
            ray.wait(
                [
                    analyzer.run.remote(
                        end_by_stopwatcher=args.end_by_stopwatcher,
                        scenarios=scenarios,
                        ego=args.ego,
                        max_episode_steps=args.max_steps,
                        policy=args.policy,
                        video_rate=args.video,
                        timestep_sec=timestep_sec,
                        custom_traci_functions=[
                            edge_lane_data_function,
                            vehicle_data_function,
                        ],
                    )
                ]
            )
            end = timeit.default_timer()
            print("Time taken:", end - start)
            ray.wait([analyzer.save_data.remote(save_dir=args.output)])
            ray.wait(
                [
                    analyzer.draw_plots.remote(
                        save_dir=args.output, failure_time=args.failure_time
                    )
                ]
            )



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import torch
from torch import nn
from torch.distributions.normal import Normal


class ActorNetwork(nn.Module):
    def __init__(
        self, state_size, action_size, hidden_units, social_feature_encoder=None
    ):
        super(ActorNetwork, self).__init__()
        self.social_feature_encoder = social_feature_encoder

        self.model = nn.Sequential(
            nn.Linear(state_size, hidden_units),
            nn.ReLU(),
            nn.Linear(hidden_units, hidden_units),
            nn.ReLU(),
            nn.Linear(hidden_units, action_size),
            nn.Tanh(),
        )

    def forward(self, state, training=False):
        low_dim_state = state["low_dim_states"]
        social_vehicles_state = state["social_vehicles"]

        aux_losses = {}
        social_feature = []
        if isinstance(social_vehicles_state, list):
            social_vehicles_state = torch.cat(social_vehicles_state)
        if self.social_feature_encoder is not None:
            social_feature, social_encoder_aux_losses = self.social_feature_encoder(
                social_vehicles_state, training
            )
            aux_losses.update(social_encoder_aux_losses)
        else:
            social_feature = [e.reshape(1, -1) for e in social_vehicles_state]
        social_feature = torch.cat(social_feature, 0) if len(social_feature) > 0 else []
        state = (
            torch.cat([low_dim_state, social_feature], -1)
            if len(social_feature) > 0
            else low_dim_state
        )
        a = self.model(state)

        if training:
            return a, aux_losses
        else:
            return a, {}


class CriticNetwork(nn.Module):
    def __init__(self, state_size, hidden_units, social_feature_encoder=None):
        super(CriticNetwork, self).__init__()
        self.social_feature_encoder = social_feature_encoder

        self.model = nn.Sequential(
            nn.Linear(state_size, hidden_units),
            nn.ReLU(),
            nn.Linear(hidden_units, hidden_units),
            nn.ReLU(),
            nn.Linear(hidden_units, 1),
        )

    def forward(self, state, training=False):
        low_dim_state = state["low_dim_states"]
        social_vehicles_state = state["social_vehicles"]
        aux_losses = {}
        social_feature = []
        if isinstance(social_vehicles_state, list):
            social_vehicles_state = torch.cat(social_vehicles_state)
        if self.social_feature_encoder is not None:
            social_feature, social_encoder_aux_losses = self.social_feature_encoder(
                social_vehicles_state, training
            )
            aux_losses.update(social_encoder_aux_losses)
        else:
            social_feature = [e.reshape(1, -1) for e in social_vehicles_state]
        social_feature = torch.cat(social_feature, 0) if len(social_feature) > 0 else []

        state = (
            torch.cat([low_dim_state, social_feature], -1)
            if len(social_feature) > 0
            else low_dim_state
        )
        q = self.model(state)
        if training:
            return q, aux_losses
        else:
            return q, {}


class PPONetwork(nn.Module):
    def __init__(
        self,
        action_size,
        state_size,
        seed=None,
        hidden_units=64,
        init_std=0.5,
        social_feature_encoder_class=None,
        social_feature_encoder_params=None,
    ):
        super(PPONetwork, self).__init__()

        if seed is not None:
            torch.manual_seed(seed)
        self.social_feature_encoder_class = social_feature_encoder_class
        self.social_feature_encoder_params = social_feature_encoder_params

        self.critic = CriticNetwork(
            state_size,
            hidden_units,
            social_feature_encoder=self.social_feature_encoder_class(
                **self.social_feature_encoder_params
            )
            if self.social_feature_encoder_class
            else None,
        )

        self.actor = ActorNetwork(
            state_size,
            action_size,
            hidden_units,
            social_feature_encoder=self.social_feature_encoder_class(
                **self.social_feature_encoder_params
            )
            if self.social_feature_encoder_class
            else None,
        )

        # self.init_last_layer(self.actor)
        self.log_std = nn.Parameter(torch.log(init_std * torch.ones(1, action_size)))

    def forward(self, x, training=False):
        value, critic_aux_loss = self.critic(x, training=training)
        mu, actor_aux_loss = self.actor(x, training=training)
        std = torch.ones_like(mu) * 0.5
        dist = Normal(mu, std)

        if training:
            aux_losses = {}
            for k, v in actor_aux_loss.items():
                aux_losses.update({"actor/{}".format(k): v})
            for k, v in critic_aux_loss.items():
                aux_losses.update({"critic/{}".format(k): v})
            return (dist, value), aux_losses
        else:
            return dist, value

    def init_last_layer(self, actor):
        """Initialize steering to zero and throttle to maximum"""
        # nn.init.constant_(actor.model[-2].weight.data[0], 1.0)
        # nn.init.constant_(actor.model[-2].weight.data[1], 0.0)
        nn.init.constant_(actor.model[-2].bias.data[0], 2.0)



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import unittest
from ultra.scenarios.common.social_vehicle_definitions import *


class SocialVehiclesTest(unittest.TestCase):
    def test_get_social_vehicles(self):
        all_types = ["default", "aggressive", "cautious", "blocker"]
        for behavior in all_types:
            vehicle = get_social_vehicle_behavior(behavior)
            self.assertTrue(vehicle is not None)
        vehicle = get_social_vehicle_behavior("Unknown")
        self.assertTrue(vehicle is None)

    def test_get_all_behaviors(self):
        all_types = ["default", "aggressive", "cautious", "blocker"]
        vehicles = get_all_behaviors()
        self.assertTrue(sorted(vehicles) == sorted(all_types))

    def test_get_social_vehicle_color(self):
        vehicles = [
            "default-2244061760535854552",
            "aggressive-224406176053585455",
            "cautious-224406176053585455",
            "blocker-224406176053585455",
            "stopwatcher_aggressive--2078631594661947501",
        ]
        for vid in vehicles:
            behavior_id = vid.split("-")[0]
            b_id, color = get_social_vehicle_color(vid)
            self.assertTrue(b_id == behavior_id)

        unknown_type = "dummy-None"
        b_id, color = get_social_vehicle_color(unknown_type)
        self.assertTrue(b_id, f" Unknown ***** {unknown_type} *******")



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import ray
from packaging import version


def default_ray_kwargs(**kwargs):
    ray_kwargs = {}
    _system_config = {
        # Needed to deal with cores that lock up for > 10 seconds
        "num_heartbeats_timeout": 10000,
        # "raylet_heartbeat_timeout_milliseconds": 10,
        # "object_timeout_milliseconds": 200,
    }
    if version.parse(ray.__version__) > version.parse("0.8"):
        ray_kwargs["_system_config"] = _system_config
    else:
        ray_kwargs["_internal_config"] = _system_config
    ray_kwargs.update(kwargs)

    return ray_kwargs


# TODO: Perhaps start cluster manually instead of `ray.init()`
# See https://github.com/ray-project/ray/blob/174bef56d452b6f86db167ecb80e7f23176079b6/python/ray/tests/conftest.py#L110



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import numpy as np
import torch
from scipy.spatial import distance
import random, math, gym
from sys import path
from collections import OrderedDict
from ultra.baselines.common.baseline_state_preprocessor import BaselineStatePreprocessor
from ultra.baselines.common.social_vehicle_config import get_social_vehicle_configs
from ultra.baselines.common.yaml_loader import load_yaml

path.append("./ultra")
from ultra.utils.common import (
    get_closest_waypoint,
    get_path_to_goal,
    ego_social_safety,
    rotate2d_vector,
)

seed = 0
random.seed(seed)
num_lookahead = 100


class BaselineAdapter:
    def __init__(self, agent_name):
        self.policy_params = load_yaml(
            f"ultra/baselines/{agent_name}/{agent_name}/params.yaml"
        )

        social_vehicle_params = self.policy_params["social_vehicles"]
        social_vehicle_params["observation_num_lookahead"] = self.policy_params[
            "observation_num_lookahead"
        ]
        self.observation_num_lookahead = social_vehicle_params[
            "observation_num_lookahead"
        ]
        self.num_social_features = social_vehicle_params["num_social_features"]
        self.social_capacity = social_vehicle_params["social_capacity"]
        self.social_vehicle_config = get_social_vehicle_configs(
            encoder_key=social_vehicle_params["encoder_key"],
            num_social_features=self.num_social_features,
            social_capacity=self.social_capacity,
            seed=social_vehicle_params["seed"],
            social_policy_hidden_units=social_vehicle_params[
                "social_policy_hidden_units"
            ],
            social_policy_init_std=social_vehicle_params["social_policy_init_std"],
        )

        self.social_vehicle_encoder = self.social_vehicle_config["encoder"]

        self.state_preprocessor = BaselineStatePreprocessor(
            social_vehicle_config=self.social_vehicle_config,
            observation_waypoints_lookahead=self.observation_num_lookahead,
            action_size=2,
        )

        self.social_feature_encoder_class = self.social_vehicle_encoder[
            "social_feature_encoder_class"
        ]
        self.social_feature_encoder_params = self.social_vehicle_encoder[
            "social_feature_encoder_params"
        ]

        self.state_size = self.state_preprocessor.num_low_dim_states
        if self.social_feature_encoder_class:
            self.state_size += self.social_feature_encoder_class(
                **self.social_feature_encoder_params
            ).output_dim
        else:
            self.state_size += self.social_capacity * self.num_social_features

    @property
    def observation_space(self):
        low_dim_states_shape = self.state_preprocessor.num_low_dim_states
        if self.social_feature_encoder_class:
            social_vehicle_shape = self.social_feature_encoder_class(
                **self.social_feature_encoder_params
            ).output_dim
        else:
            social_vehicle_shape = self.social_capacity * self.num_social_features
        return gym.spaces.Dict(
            {
                "low_dim_states": gym.spaces.Box(
                    low=-1e10,
                    high=1e10,
                    shape=(low_dim_states_shape,),
                    dtype=torch.Tensor,
                ),
                "social_vehicles": gym.spaces.Box(
                    low=-1e10,
                    high=1e10,
                    shape=(self.social_capacity, self.num_social_features),
                    dtype=torch.Tensor,
                ),
            }
        )

    @property
    def action_space(self):
        return gym.spaces.Box(
            low=np.array([0.0, 0.0, -1.0]),
            high=np.array([1.0, 1.0, 1.0]),
            dtype=np.float32,
        )

    # def action_adapter(self, model_action):
    #     # print why this doesn't go through?
    #     throttle, brake, steering = model_action
    #     # print(M)
    #     return np.array([throttle, brake, steering * np.pi * 0.25])

    def observation_adapter(self, env_observation):
        state = self.state_preprocessor(
            state=env_observation,
            observation_num_lookahead=self.observation_num_lookahead,
            social_capacity=self.social_capacity,
            social_vehicle_config=self.social_vehicle_config,
            # prev_action=self.prev_action
        )

        if len(state["social_vehicles"]) < self.social_capacity:
            remain = self.social_capacity - len(state["social_vehicles"])
            empty_social_vehicles = np.zeros(shape=(remain, 4), dtype=np.float32)
            state["social_vehicles"] = np.concatenate(
                (state["social_vehicles"], empty_social_vehicles)
            )
        # todo would this cause any issues for precog
        state["social_vehicles"] = state["social_vehicles"][: self.social_capacity]
        return state  # ego=ego, env_observation=env_observation)

    @staticmethod
    def reward_adapter(observation, reward):
        env_reward = reward
        ego_events = observation.events
        ego_observation = observation.ego_vehicle_state
        start = observation.ego_vehicle_state.mission.start
        goal = observation.ego_vehicle_state.mission.goal
        path = get_path_to_goal(
            goal=goal, paths=observation.waypoint_paths, start=start
        )

        linear_jerk = np.linalg.norm(ego_observation.linear_jerk)
        angular_jerk = np.linalg.norm(ego_observation.angular_jerk)

        # Distance to goal
        ego_2d_position = ego_observation.position[0:2]
        goal_dist = distance.euclidean(ego_2d_position, goal.position)

        closest_wp, _ = get_closest_waypoint(
            num_lookahead=num_lookahead,
            goal_path=path,
            ego_position=ego_observation.position,
            ego_heading=ego_observation.heading,
        )
        angle_error = closest_wp.relative_heading(
            ego_observation.heading
        )  # relative heading radians [-pi, pi]

        # Distance from center
        signed_dist_from_center = closest_wp.signed_lateral_error(
            observation.ego_vehicle_state.position
        )
        lane_width = closest_wp.lane_width * 0.5
        ego_dist_center = signed_dist_from_center / lane_width

        # number of violations
        (ego_num_violations, social_num_violations,) = ego_social_safety(
            observation,
            d_min_ego=1.0,
            t_c_ego=1.0,
            d_min_social=1.0,
            t_c_social=1.0,
            ignore_vehicle_behind=True,
        )

        speed_fraction = max(0, ego_observation.speed / closest_wp.speed_limit)
        ego_step_reward = 0.02 * min(speed_fraction, 1) * np.cos(angle_error)
        ego_speed_reward = min(
            0, (closest_wp.speed_limit - ego_observation.speed) * 0.01
        )  # m/s
        ego_collision = len(ego_events.collisions) > 0
        ego_collision_reward = -1.0 if ego_collision else 0.0
        ego_off_road_reward = -1.0 if ego_events.off_road else 0.0
        ego_off_route_reward = -1.0 if ego_events.off_route else 0.0
        ego_wrong_way = -0.02 if ego_events.wrong_way else 0.0
        ego_goal_reward = 0.0
        ego_time_out = 0.0
        ego_dist_center_reward = -0.002 * min(1, abs(ego_dist_center))
        ego_angle_error_reward = -0.005 * max(0, np.cos(angle_error))
        ego_reached_goal = 1.0 if ego_events.reached_goal else 0.0
        ego_safety_reward = -0.02 if ego_num_violations > 0 else 0
        social_safety_reward = -0.02 if social_num_violations > 0 else 0
        ego_lat_speed = 0.0  # -0.1 * abs(long_lat_speed[1])
        ego_linear_jerk = -0.0001 * linear_jerk
        ego_angular_jerk = -0.0001 * angular_jerk * math.cos(angle_error)
        env_reward /= 100
        # DG: Different speed reward
        ego_speed_reward = -0.1 if speed_fraction >= 1 else 0.0
        ego_speed_reward += -0.01 if speed_fraction < 0.01 else 0.0

        rewards = sum(
            [
                ego_goal_reward,
                ego_collision_reward,
                ego_off_road_reward,
                ego_off_route_reward,
                ego_wrong_way,
                ego_speed_reward,
                # ego_time_out,
                ego_dist_center_reward,
                ego_angle_error_reward,
                ego_reached_goal,
                ego_step_reward,
                env_reward,
                # ego_linear_jerk,
                # ego_angular_jerk,
                # ego_lat_speed,
                # ego_safety_reward,
                # social_safety_reward,
            ]
        )
        return rewards



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import json
import os
import sys

from ultra.utils.ray import default_ray_kwargs

# Set environment to better support Ray
os.environ["MKL_NUM_THREADS"] = "1"
import argparse
import pickle
import time

import dill
import gym
import psutil
import ray
import torch

from smarts.zoo.registry import make
from ultra.evaluate import evaluation_check
from ultra.utils.episode import episodes

num_gpus = 1 if torch.cuda.is_available() else 0


# @ray.remote(num_gpus=num_gpus / 2, max_calls=1)
@ray.remote(num_gpus=num_gpus / 2)
def train(
    scenario_info,
    num_episodes,
    policy_classes,
    max_episode_steps,
    eval_info,
    timestep_sec,
    headless,
    seed,
    log_dir,
    policy_ids=None,
):
    torch.set_num_threads(1)
    total_step = 0
    finished = False

    # Make agent_ids in the form of 000, 001, ..., 010, 011, ..., 999, 1000, ...;
    # or use the provided policy_ids if available.
    agent_ids = (
        ["0" * max(0, 3 - len(str(i))) + str(i) for i in range(len(policy_classes))]
        if not policy_ids
        else policy_ids
    )
    # Ensure there is an ID for each policy, and a policy for each ID.
    assert len(agent_ids) == len(policy_classes), (
        "The number of agent IDs provided ({}) must be equal to "
        "the number of policy classes provided ({}).".format(
            len(agent_ids), len(policy_classes)
        )
    )

    # Assign the policy classes to their associated ID.
    agent_classes = {
        agent_id: policy_class
        for agent_id, policy_class in zip(agent_ids, policy_classes)
    }
    # Create the agent specifications matched with their associated ID.
    agent_specs = {
        agent_id: make(locator=policy_class, max_episode_steps=max_episode_steps)
        for agent_id, policy_class in agent_classes.items()
    }
    # Create the agents matched with their associated ID.
    agents = {
        agent_id: agent_spec.build_agent()
        for agent_id, agent_spec in agent_specs.items()
    }

    # Create the environment.
    env = gym.make(
        "ultra.env:ultra-v0",
        agent_specs=agent_specs,
        scenario_info=scenario_info,
        headless=headless,
        timestep_sec=timestep_sec,
        seed=seed,
    )

    # Define an 'etag' for this experiment's data directory based off policy_classes.
    # E.g. From a ["ultra.baselines.dqn:dqn-v0", "ultra.baselines.ppo:ppo-v0"]
    # policy_classes list, transform it to an etag of "dqn-v0:ppo-v0".
    etag = ":".join([policy_class.split(":")[-1] for policy_class in policy_classes])

    for episode in episodes(num_episodes, etag=etag, log_dir=log_dir):
        # Reset the environment and retrieve the initial observations.
        observations = env.reset()
        dones = {"__all__": False}
        infos = None
        episode.reset()
        experiment_dir = episode.experiment_dir

        # Save relevant agent metadata.
        if not os.path.exists(f"{experiment_dir}/agent_metadata.pkl"):
            if not os.path.exists(experiment_dir):
                os.makedirs(experiment_dir)
            with open(f"{experiment_dir}/agent_metadata.pkl", "wb") as metadata_file:
                dill.dump(
                    {
                        "agent_ids": agent_ids,
                        "agent_classes": agent_classes,
                        "agent_specs": agent_specs,
                    },
                    metadata_file,
                    pickle.HIGHEST_PROTOCOL,
                )

        while not dones["__all__"]:
            # Break if any of the agent's step counts is 1000000 or greater.
            if any([episode.get_itr(agent_id) >= 1000000 for agent_id in agents]):
                finished = True
                break

            # Perform the evaluation check.
            evaluation_check(
                agents=agents,
                agent_ids=agent_ids,
                policy_classes=agent_classes,
                episode=episode,
                log_dir=log_dir,
                max_episode_steps=max_episode_steps,
                **eval_info,
                **env.info,
            )

            # Request and perform actions on each agent that received an observation.
            actions = {
                agent_id: agents[agent_id].act(observation, explore=True)
                for agent_id, observation in observations.items()
            }
            next_observations, rewards, dones, infos = env.step(actions)

            # Active agents are those that receive observations in this step and the next
            # step. Step each active agent (obtaining their network loss if applicable).
            active_agent_ids = observations.keys() & next_observations.keys()
            loss_outputs = {
                agent_id: agents[agent_id].step(
                    state=observations[agent_id],
                    action=actions[agent_id],
                    reward=rewards[agent_id],
                    next_state=next_observations[agent_id],
                    done=dones[agent_id],
                    info=infos[agent_id],
                )
                for agent_id in active_agent_ids
            }

            # Record the data from this episode.
            episode.record_step(
                agent_ids_to_record=active_agent_ids,
                infos=infos,
                rewards=rewards,
                total_step=total_step,
                loss_outputs=loss_outputs,
            )

            # Update variables for the next step.
            total_step += 1
            observations = next_observations

        # Normalize the data and record this episode on tensorboard.
        episode.record_episode()
        episode.record_tensorboard()

        if finished:
            break

    env.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser("intersection-training")
    parser.add_argument(
        "--task", help="Tasks available : [0, 1, 2]", type=str, default="1"
    )
    parser.add_argument(
        "--level",
        help="Levels available : [easy, medium, hard, no-traffic]",
        type=str,
        default="easy",
    )
    parser.add_argument(
        "--policy",
        help="Policies available : [ppo, sac, td3, dqn, bdqn]",
        type=str,
        default="sac",
    )
    parser.add_argument(
        "--episodes", help="Number of training episodes", type=int, default=1000000
    )
    parser.add_argument(
        "--max-episode-steps",
        help="Maximum number of steps per episode",
        type=int,
        default=10000,
    )
    parser.add_argument(
        "--timestep", help="Environment timestep (sec)", type=float, default=0.1
    )
    parser.add_argument(
        "--headless", help="Run without envision", type=bool, default=False
    )
    parser.add_argument(
        "--eval-episodes", help="Number of evaluation episodes", type=int, default=200
    )
    parser.add_argument(
        "--eval-rate",
        help="Evaluation rate based on number of observations",
        type=int,
        default=10000,
    )
    parser.add_argument(
        "--seed",
        help="Environment seed",
        default=2,
        type=int,
    )
    parser.add_argument(
        "--log-dir",
        help="Log directory location",
        default="logs",
        type=str,
    )
    parser.add_argument(
        "--policy-ids",
        help="Name of each specified policy",
        default=None,
        type=str,
    )

    base_dir = os.path.dirname(__file__)
    pool_path = os.path.join(base_dir, "agent_pool.json")
    args = parser.parse_args()

    # Obtain the policy class strings for each specified policy.
    policy_classes = []
    with open(pool_path, "r") as f:
        data = json.load(f)
        for policy in args.policy.split(","):
            if policy in data["agents"].keys():
                policy_classes.append(
                    data["agents"][policy]["path"]
                    + ":"
                    + data["agents"][policy]["locator"]
                )
            else:
                raise ImportError("Invalid policy name. Please try again")

    # Obtain the policy class IDs from the arguments.
    policy_ids = args.policy_ids.split(",") if args.policy_ids else None

    ray.init()
    ray.wait(
        [
            train.remote(
                scenario_info=(args.task, args.level),
                num_episodes=int(args.episodes),
                max_episode_steps=int(args.max_episode_steps),
                eval_info={
                    "eval_rate": float(args.eval_rate),
                    "eval_episodes": int(args.eval_episodes),
                },
                timestep_sec=float(args.timestep),
                headless=args.headless,
                policy_classes=policy_classes,
                seed=args.seed,
                log_dir=args.log_dir,
                policy_ids=policy_ids,
            )
        ]
    )



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import numpy as np


def basic_vehicle_initial_distribution(probability, min_step=100000, extra_args=None):

    count = 1
    while True:
        p = np.random.uniform(0, 1) < probability
        if probability == 0.0 or p or count > min_step:
            break
        count += 1
    return count


def basic_begin_time_init_func(num_lanes, num_vehicle, **params):
    begin_times = [[] for _ in range(num_lanes)]
    for j in range(num_lanes):
        [
            begin_times[j].append(basic_vehicle_initial_distribution(**params))
            for _ in range(num_vehicle)
        ]
        begin_times[j] = np.cumsum(begin_times[j])
        begin_times[j] = [float(e) for e in begin_times[j]]
    return begin_times


def burst_begin_time_init_func(
    num_lanes,
    num_vehicle,
    vehicle_cluster_size=(4, 5),
    time_between_cluster=(12, 25),
    time_for_each_cluster=5,
):
    begin_times = []
    for j in range(num_lanes):
        vehicle_left = num_vehicle
        _begin_times = []
        time_so_far = 0

        while vehicle_left > 0:
            num_vehicle_in_cluster = min(
                np.random.randint(vehicle_cluster_size[0], vehicle_cluster_size[1] + 1),
                vehicle_left,
            )

            vehicle_spawn_time = np.random.uniform(
                0, time_for_each_cluster, size=num_vehicle_in_cluster
            )
            # vehicle_spawn_time = np.clip(np.random.exponential(10, size=len(behaviors)), 0.0, allocated_time_for_cluster)
            vehicle_spawn_time += time_so_far
            _begin_times.extend(vehicle_spawn_time)

            time_to_next_cluster = np.random.uniform(*time_between_cluster)
            time_so_far += time_to_next_cluster

            vehicle_left -= num_vehicle_in_cluster

        begin_times.append(_begin_times)
    return begin_times


def poisson_init(num_lanes, num_vehicle, temperature=10):
    vehicle_spawn_time = np.random.exponential(
        temperature, size=(num_lanes, num_vehicle)
    )
    vehicle_spawn_time = list(vehicle_spawn_time)
    vehicle_spawn_time = [np.cumsum(e) for e in vehicle_spawn_time]
    return vehicle_spawn_time



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import collections.abc
import numpy as np
import torch

from ultra.baselines.common.state_preprocessor import StatePreprocessor
from ultra.utils.common import rotate2d_vector


class BaselineStatePreprocessor(StatePreprocessor):
    """The State Preprocessor used by the baseline agents."""

    _NORMALIZATION_VALUES = {
        "speed": 30.0,
        "distance_from_center": 1.0,
        "steering": 3.14,  # radians
        "angle_error": 3.14,  # radians
        "relative_goal_position": 100.0,
        # "action": 1.0,  # 2
        "waypoints_lookahead": 10.0,
        "road_speed": 30.0,
    }

    def __init__(
        self,
        social_vehicle_config,
        observation_waypoints_lookahead,
        action_size,
    ):
        self._state_description = self.get_state_description(
            social_vehicle_config, observation_waypoints_lookahead, action_size
        )

    @staticmethod
    def get_state_description(
        social_vehicle_config, observation_waypoints_lookahead, action_size
    ):
        return {
            "low_dim_states": {
                "speed": 1,
                "distance_from_center": 1,
                "steering": 1,
                "angle_error": 1,
                "relative_goal_position": 2,
                # "action": int(action_size),  # 2
                "waypoints_lookahead": 2 * int(observation_waypoints_lookahead),
                "road_speed": 1,
            },
            "social_vehicles": int(social_vehicle_config["num_social_features"])
            if int(social_vehicle_config["social_capacity"]) > 0
            else 0,
        }

    @property
    def num_low_dim_states(self):
        return sum(self._state_description["low_dim_states"].values())

    def _preprocess_state(
        self,
        state,
        observation_num_lookahead,
        social_capacity,
        social_vehicle_config,
    ):
        state = self._adapt_observation_for_baseline(state)

        # Obtain the next waypoints.
        _, lookahead_waypoints = self.extract_closest_waypoint(
            ego_goal_path=state["goal_path"],
            ego_position=state["ego_position"],
            ego_heading=state["heading"],
            num_lookahead=observation_num_lookahead,
        )
        state["waypoints_lookahead"] = np.hstack(lookahead_waypoints)

        # Normalize states and concatenate.
        normalized = [
            self._normalize(key, state[key])
            for key in self._state_description["low_dim_states"]
        ]
        low_dim_states = [
            value
            if isinstance(value, collections.abc.Iterable)
            else np.asarray([value]).astype(np.float32)
            for value in normalized
        ]
        low_dim_states = torch.cat(
            [torch.from_numpy(e).float() for e in low_dim_states], dim=-1
        )

        # Apply the social vehicle encoder (if applicable).
        state["social_vehicles"] = (
            self.get_social_vehicles_representation(
                social_vehicles=state["social_vehicles"],
                social_vehicle_config=social_vehicle_config,
                ego_position=state["ego_position"],
                ego_heading=state["heading"],
                ego_waypoints=state["waypoint_paths"],
            )
            if social_capacity > 0
            else []
        )

        # Check if any social capacity is 0.
        social_vehicle_dimension = self._state_description["social_vehicles"]
        social_vehicles = torch.empty(0, 0)
        if social_vehicle_dimension:
            social_vehicles = torch.from_numpy(
                np.asarray(state["social_vehicles"])
            ).float()
            social_vehicles = social_vehicles.reshape((-1, social_vehicle_dimension))

        out = {
            "low_dim_states": low_dim_states.numpy(),
            "social_vehicles": social_vehicles.numpy(),
        }
        return out

    def _adapt_observation_for_baseline(self, state):
        # Get basic information about the ego vehicle.
        ego_position = self.extract_ego_position(state)
        ego_heading = self.extract_ego_heading(state)
        ego_speed = self.extract_ego_speed(state)
        ego_steering = self.extract_ego_steering(state)
        ego_start = self.extract_ego_start(state)
        ego_goal = self.extract_ego_goal(state)
        ego_waypoints = self.extract_ego_waypoints(state)
        social_vehicle_states = self.extract_social_vehicles(state)

        # Identify the path the ego is following.
        ego_goal_path = self.extract_ego_goal_path(
            ego_goal=ego_goal,
            ego_waypoints=ego_waypoints,
            ego_start=ego_start,
        )

        # Get the closest waypoint to the ego.
        ego_closest_waypoint, _ = self.extract_closest_waypoint(
            ego_goal_path=ego_goal_path,
            ego_position=ego_position,
            ego_heading=ego_heading,
        )

        # Calculate the ego's distance from the center of the lane.
        signed_distance_from_center = ego_closest_waypoint.signed_lateral_error(
            ego_position
        )
        lane_width = ego_closest_waypoint.lane_width * 0.5
        ego_distance_from_center = signed_distance_from_center / lane_width

        # Calculate the ego's relative, rotated position from the goal.
        ego_relative_rotated_goal_position = rotate2d_vector(
            np.asarray(ego_goal.position[0:2]) - np.asarray(ego_position[0:2]),
            -ego_heading,
        )

        basic_state = dict(
            speed=ego_speed,
            relative_goal_position=ego_relative_rotated_goal_position,
            distance_from_center=ego_distance_from_center,
            steering=ego_steering,
            angle_error=ego_closest_waypoint.relative_heading(ego_heading),
            social_vehicles=social_vehicle_states,
            road_speed=ego_closest_waypoint.speed_limit,
            start=ego_start.position,
            goal=ego_goal.position,
            heading=ego_heading,
            goal_path=ego_goal_path,
            ego_position=ego_position,
            waypoint_paths=ego_waypoints,
            events=state.events,
        )
        return basic_state

    def _normalize(self, key, value):
        if key not in self._NORMALIZATION_VALUES:
            return value
        return value / self._NORMALIZATION_VALUES[key]



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from ultra.scenarios.common.begin_time_init_funcs import *

# -----------------------------------------
#             cheat sheet
# -----------------------------------------
# 'north-south': ["north-NS", "south-NS"]
# 'south-north': ["south-SN", "north-SN"]
# 'south-east':  ["south-SN", "east-WE"]
# 'south-west':  ["south-SN", "west-EW"]
# 'north-east':  ["north-NS", "east-WE"]
# 'north-west':  ["north-NS", "west-EW"]
# 'east-north':  ["east-EW", "north-SN"]
# 'east-south': ["east-EW", "south-NS"]
# 'east-west':  ["east-EW", "west-EW"]
# 'west-north': ["west-WE", "north-SN"]
# 'west-south': ["west-WE", "south-NS"]
# 'west-east':  ["west-WE", "east-WE"]
# -----------------------------------------


prob_easy = 0.02
prob_medium = 0.04
prob_heavy = 0.06
behavior_distribution = {
    "default": 0.85,
    "aggressive": 0.05,
    "cautious": 0.05,
    "blocker": 0.05,
}
t_patterns = {
    "no-traffic": {  # t-intersection
        "routes": {
            "south-west": {
                "vehicles": 2,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_easy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
            "south-east": None,
            "south-north": None,  # blocking
            "west-east": None,
            "west-north": None,  # blocking
            "west-south": None,
            "east-west": None,
            "east-north": None,  # blocking
            "east-south": {
                "vehicles": 2,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_easy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
        },
        # t-intersection has no north route
    },
    "low-density": {  # t-intersection
        "routes": {
            "south-west": {
                "vehicles": 100,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_easy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
            "south-east": {
                "vehicles": 100,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_easy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
            "south-north": None,  # blocking
            "west-east": {
                "vehicles": 1000,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_easy},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
            "west-north": None,  # blocking
            "west-south": {
                "vehicles": 100,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_easy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
            "east-west": {
                "vehicles": 1000,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_easy},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
            "east-north": None,  # blocking
            "east-south": {
                "vehicles": 100,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_easy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
        },
        "ego_hijacking_params": {
            "zone_range": [5, 10],
            "wait_to_hijack_limit_s": 10,
            "start_time": "default",  # any value or default for LANE_LENGTH / speed_m_per_s
        },
        # t-intersection has no north route
    },
    "mid-density": {  # t-intersection
        "routes": {
            "south-west": {
                "vehicles": 100,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_medium},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
            "south-east": {
                "vehicles": 100,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_medium},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
            "south-north": None,  # blocking
            "west-east": {
                "vehicles": 1000,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_medium},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
            "west-north": None,  # blocking
            "east-west": {
                "vehicles": 1000,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_medium},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
            "east-north": None,  # blocking
            "east-south": {
                "vehicles": 100,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_medium},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
            "west-south": {
                "vehicles": 100,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_medium},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
        },
        "ego_hijacking_params": {
            "zone_range": [5, 10],
            "wait_to_hijack_limit_s": 10,
            "start_time": "default",  # any value or default for LANE_LENGTH / speed_m_per_s
        },
    },
    "high-density": {  # t-intersection
        "routes": {
            "south-west": {
                "vehicles": 100,
                "distribution": behavior_distribution,
                "start_end_on_different_lanes_probability": 0.0,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_heavy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },  # blocking
            "south-east": {
                "vehicles": 100,
                "distribution": behavior_distribution,
                "start_end_on_different_lanes_probability": 0.0,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_heavy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },  # blocking
            "south-north": None,  # blocking
            "west-east": {
                "vehicles": 1000,
                "distribution": behavior_distribution,
                "start_end_on_different_lanes_probability": 0.0,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_heavy},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
            "west-north": None,  # blocking
            "west-south": {
                "vehicles": 100,
                "distribution": behavior_distribution,
                "start_end_on_different_lanes_probability": 0.0,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_heavy},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },  # blocking
            "east-west": {
                "vehicles": 1000,
                "distribution": behavior_distribution,
                "start_end_on_different_lanes_probability": 0.0,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_heavy},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
            "east-north": None,  # blocking
            "east-south": {
                "vehicles": 100,
                "distribution": behavior_distribution,
                "start_end_on_different_lanes_probability": 0.0,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_heavy},
                },
                "has_turn": True,
                "deadlock_optimization": False,
            },
            # t-intersection has no north route
        },
        "ego_hijacking_params": {
            "zone_range": [5, 10],
            "wait_to_hijack_limit_s": 10,
            "start_time": "default",  # any value or default for LANE_LENGTH / speed_m_per_s
        },
    },
    "p-stopwatchers": {  # t-intersection
        "routes": {
            "south-west": None,
            "south-east": None,
            "south-north": None,  # blocking
            "west-east": None,
            "west-north": None,  # blocking
            "west-south": None,
            "east-south": None,
            "east-north": None,  # blocking
            "east-west": {
                "vehicles": 10,
                "distribution": {"default": 0.4, "aggressive": 0.4, "cautious": 0.1},
                "start_end_on_different_lanes_probability": 1,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": 0.0},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
        }
    },
    "p-test": {  # t-intersection
        "routes": {
            "south-west": {
                "vehicles": 100,
                "distribution": {"default": 0.4, "aggressive": 0.4, "cautious": 0.1},
                "start_end_on_different_lanes_probability": 1,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": 0.0},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
            "south-east": {
                "vehicles": 100,
                "distribution": {"default": 0.4, "aggressive": 0.4, "cautious": 0.1},
                "start_end_on_different_lanes_probability": 1,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": 0.0},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
            "south-north": None,  # blocking
            "west-east": None,
            "west-north": None,  # blocking
            "west-south": None,
            "east-south": None,
            "east-north": None,  # blocking
            "east-west": {
                "vehicles": 100,
                "distribution": {"default": 1},
                "start_end_on_different_lanes_probability": 1,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": 0.0},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
        },
        "ego_hijacking_params": {
            "zone_range": [5, 10],
            "wait_to_hijack_limit_s": 100,
            "start_time": "default",  # any value or default for LANE_LENGTH / speed_m_per_s
        },
    },
}

cross_patterns = {
    "no-traffic": {  # t-intersection
        "routes": {
            "south-west": {
                "vehicles": 2,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_easy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
            "south-east": None,
            "south-north": None,  # blocking
            "west-east": None,
            "west-north": None,  # blocking
            "west-south": None,
            "east-west": None,
            "east-north": None,  # blocking
            "east-south": {
                "vehicles": 2,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_easy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
            "north-south": None,
            "north-east": None,
            "north-west": None,
        },
        # t-intersection has no north route
    },
    "low-density": {  # t-intersection
        "routes": {
            "south-west": {
                "vehicles": 100,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_easy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
            "south-east": {
                "vehicles": 100,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_easy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
            "south-north": {
                "vehicles": 1000,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_easy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
            "west-east": None,
            "west-north": None,  # blocking
            "west-south": None,
            "east-west": None,
            "east-north": None,  # blocking
            "east-south": None,
            "north-south": {
                "vehicles": 1000,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_easy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
            "north-west": {
                "vehicles": 100,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_easy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
            "north-east": {
                "vehicles": 100,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_easy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
        },
        "ego_hijacking_params": {
            "zone_range": [5, 10],
            "wait_to_hijack_limit_s": 10,
            "start_time": "default",  # any value or default for LANE_LENGTH / speed_m_per_s
        },
    },
    "mid-density": {  # t-intersection
        "routes": {
            "south-west": {
                "vehicles": 100,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_medium},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
            "south-east": {
                "vehicles": 100,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_medium},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
            "south-north": {
                "vehicles": 1000,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_medium},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
            "west-east": None,
            "west-north": None,  # blocking
            "east-west": None,
            "east-north": None,  # blocking
            "east-south": None,
            "west-south": None,
            "north-south": {
                "vehicles": 1000,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_medium},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
            "north-east": {
                "vehicles": 100,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_medium},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
            "north-west": {
                "vehicles": 100,
                "start_end_on_different_lanes_probability": 0.0,
                "distribution": behavior_distribution,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_medium},
                },
                "has_turn": False,
                "deadlock_optimization": False,
            },
        },
        "ego_hijacking_params": {
            "zone_range": [5, 10],
            "wait_to_hijack_limit_s": 10,
            "start_time": "default",  # any value or default for LANE_LENGTH / speed_m_per_s
        },
    },
    "high-density": {  # t-intersection
        "routes": {
            "south-west": {
                "vehicles": 100,
                "distribution": behavior_distribution,
                "start_end_on_different_lanes_probability": 0.0,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_heavy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },  # blocking
            "south-east": {
                "vehicles": 100,
                "distribution": behavior_distribution,
                "start_end_on_different_lanes_probability": 0.0,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_heavy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },  # blocking
            "south-north": {
                "vehicles": 1000,
                "distribution": behavior_distribution,
                "start_end_on_different_lanes_probability": 0.0,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_heavy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
            "west-east": None,
            "west-north": None,  # blocking
            "west-south": None,  # blocking
            "east-west": None,
            "east-north": None,  # blocking
            "east-south": None,
            "north-south": {
                "vehicles": 1000,
                "distribution": behavior_distribution,
                "start_end_on_different_lanes_probability": 0.0,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_heavy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
            "north-east": {
                "vehicles": 100,
                "distribution": behavior_distribution,
                "start_end_on_different_lanes_probability": 0.0,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_heavy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
            "north-west": {
                "vehicles": 100,
                "distribution": behavior_distribution,
                "start_end_on_different_lanes_probability": 0.0,
                "begin_time_init": {
                    "func": basic_begin_time_init_func,
                    "params": {"probability": prob_heavy},
                },
                "has_turn": True,
                "deadlock_optimization": True,
            },
        },
        "ego_hijacking_params": {
            "zone_range": [5, 10],
            "wait_to_hijack_limit_s": 10,
            "start_time": "default",  # any value or default for LANE_LENGTH / speed_m_per_s
        },
    },
}


def get_pattern(idx, intersection_type):
    if intersection_type[-2:] == "_t":
        patterns = t_patterns
    elif intersection_type[-2:] == "_c":
        patterns = cross_patterns
    else:
        raise ("Intersection type not detected.")
    if idx in patterns:
        return patterns[idx]



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import numpy as np
import torch
from ultra.baselines.configs import Config
from ultra.baselines.common.replay_buffer import ReplayBuffer


class TD3Config(Config):
    def __init__(self, task):
        super().__init__(task=task)
        self.set_config(
            seed=2,
            social_capacity=5,
            action_size=2,
            social_vehicle_encoder="pointnet_encoder",
            save_codes=[
                "ultra/src/train.py",
                "ultra/baselines/td3/config.py",
                "ultra/baselines/td3/policy.py",
                "ultra/baselines/td3/fc_model.py",
                "ultra/utils/common.py",
                "ultra/src/adapter.py",
            ],
        )
        self.set_config(
            policy_params={
                "state_size": self.state_size,
                "action_size": self.action_size,
                "action_range": np.asarray(
                    [[-1.0, 1.0], [-1.0, 1.0]], dtype=np.float32
                ),
                "state_preprocessor": self.state_preprocessor,
                "update_rate": 5,
                "policy_delay": 2,
                "noise_clip": 0.5,
                "policy_noise": 0.2,
                "warmup": 10000,
                "actor_lr": 1e-4,
                "critic_lr": 1e-3,
                "critic_wd": 0.0,
                "actor_wd": 0.0,
                "critic_tau": 0.01,
                "actor_tau": 0.01,
                "device_name": self.device_name,
                "seed": self.seed,
                "gamma": 0.99,
                "batch_size": 128,
                "sigma": 0.3,
                "theta": 0.15,
                "dt": 1e-2,
                "replay": ReplayBuffer(
                    buffer_size=int(1e6),
                    batch_size=128,
                    state_preprocessor=self.state_preprocessor,
                    device_name=self.device_name,
                ),
                "social_feature_encoder_class": self.social_feature_encoder_class,
                "social_feature_encoder_params": self.social_feature_encoder_params,
            },
        )
